# Copyright 2015-2018 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import binascii
import collections
import hashlib
import io
import ipaddress
import json
import os
import re
import socket
import subprocess
import sys
import time
from distutils.version import LooseVersion

from charmhelpers import context
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.host import is_container
from charmhelpers.core.templating import render

import charms.promreg
from charms import apt
from charms.layer import snap
from charms.reactive import (
    clear_flag,
    endpoint_from_flag,
    helpers,
    hook,
    set_flag,
    toggle_flag,
    when,
    when_not,
)
from charms.reactive.bus import get_states

from files.telegraf_exec_metrics import METRICS

from jinja2 import Environment, FileSystemLoader, Template, exceptions

import netifaces

import yaml


DEB_BASE_DIR = "/etc/telegraf"
SNAP_BASE_DIR = "/var/snap/telegraf/current"
SUDOERS_DIR = "/etc/sudoers.d"

CONFIG_FILE = "telegraf.conf"

CONFIG_DIR = "telegraf.d"

GRAFANA_DASHBOARD_TELEGRAF_FILE_NAME = "Telegraf.json.j2"

GRAFANA_DASHBOARD_NAME = "telegraf"

SNAP_SERVICE = "snap.telegraf.telegraf"
DEB_SERVICE = "telegraf"

SNAP_OWNER = "snap_daemon"
SNAP_GROUP = "snap_daemon"
SNAP_USER = "snap_daemon"
DEB_OWNER = "root"
DEB_GROUP = "telegraf"
DEB_USER = "telegraf"

# Utilities #


class InvalidInstallMethod(Exception):
    pass


class InvalidPrometheusIPRange(Exception):
    pass


def write_telegraf_file(path, content):
    return host.write_file(
        path,
        content.encode("UTF-8"),
        owner=get_telegraf_owner(),
        group=get_telegraf_group(),
        perms=0o640,
    )


def get_install_method():
    config = hookenv.config()

    if config["install_method"] in ["deb", "snap"]:
        return config["install_method"]
    else:
        hookenv.log(
            "Invalid install_method for telegraf: {}".format(config["install_method"]),
            level=hookenv.ERROR,
        )
        raise InvalidInstallMethod()


def render_sudoers_file(filename):
    """Generate and install a sudoers file from a template to /etc/sudoers.d ."""
    template = "{}.tmpl".format(filename)
    context = {"telegraf_user": get_telegraf_user()}
    hookenv.log("Installing sudoers file {}".format(filename), level="DEBUG")
    render(
        source=template,
        templates_dir=os.path.join(get_templates_dir(), "sudoers"),
        target=os.path.join(SUDOERS_DIR, filename),
        context=context,
        perms=0o640,
    )


def remove_sudoers_file(filename):
    """Remove a sudoers file. Do nothing if the file doesn't exist."""
    path = os.path.join(SUDOERS_DIR, filename)
    if os.path.isfile(path):
        os.unlink(path)


def get_telegraf_user():
    install_method = get_install_method()

    # get_install_method() already checked that we got a valid value
    if install_method == "snap":
        return SNAP_USER
    return DEB_USER


def get_telegraf_owner():
    install_method = get_install_method()

    # get_install_method() already checked that we got a valid value
    if install_method == "snap":
        return SNAP_OWNER
    return DEB_OWNER


def get_telegraf_group():
    install_method = get_install_method()

    # get_install_method() already checked that we got a valid value
    if install_method == "snap":
        return SNAP_GROUP
    return DEB_GROUP


def get_base_dir():
    config = hookenv.config()

    if config["install_method"] == "deb":
        return DEB_BASE_DIR
    elif config["install_method"] == "snap":
        return SNAP_BASE_DIR
    else:
        hookenv.log(
            "Invalid install_method for telegraf: {}".format(config["install_method"]),
            level=hookenv.ERROR,
        )
        raise InvalidInstallMethod()


def get_service():
    config = hookenv.config()

    if config["install_method"] == "deb":
        return DEB_SERVICE
    elif config["install_method"] == "snap":
        return SNAP_SERVICE
    else:
        hookenv.log(
            "Invalid install_method for telegraf: {}".format(config["install_method"]),
            level=hookenv.ERROR,
        )
        raise InvalidInstallMethod()


def get_templates_dir():
    return os.path.join(hookenv.charm_dir(), "templates")


def get_main_config_path():
    return os.path.join(get_base_dir(), CONFIG_FILE)


def get_configs_dir():
    return os.path.join(get_base_dir(), CONFIG_DIR)


def get_files_dir():
    return os.path.join(hookenv.charm_dir(), "files")


def list_supported_plugins():
    return [k for k in hookenv.metadata()["requires"].keys() if k != "juju-info"] + [
        k for k in hookenv.metadata()["provides"].keys() if k != "juju-info"
    ]


def list_config_files():
    config_files = [get_main_config_path()]
    # only include config files for configured plugins
    current_states = get_states()

    for plugin in list_supported_plugins():
        # The prometheus_client plugin can be configured either from a relation or
        # from the juju config
        if ("plugins.{}.configured".format(plugin) in current_states.keys()) or (
            "{}.configured".format(plugin) in current_states.keys()
        ):
            # The "prometheus-client" relation sets the "prometheus_client" plugin
            if plugin == "prometheus-client":
                plugin = "prometheus_client"
            config_path = "{}/{}.conf".format(get_configs_dir(), plugin)
            config_files.append(config_path)

    if "extra_plugins.configured" in current_states.keys():
        config_files.append("{}/extra_plugins.conf".format(get_configs_dir()))
    config_files.append("{}/socket_listener.conf".format(get_configs_dir()))

    return config_files


def get_hostname_label():
    config = hookenv.config()
    hostname_fmt = config["hostname"]
    unit = get_remote_unit_name().replace("/", "-")  # / is invalid in labels.

    if hostname_fmt == "UNIT_NAME":  # Deprecated
        return unit
    env = os.environ
    model = env.get("JUJU_ENV_NAME") or env.get("JUJU_MODEL_NAME", "")
    uuid = env.get("JUJU_ENV_UUID") or env.get("JUJU_MODEL_UUID", "")
    syshost = socket.gethostname()

    return hostname_fmt.format(unit=unit, model=model, uuid=uuid, host=syshost)


def get_remote_unit_name():
    unit = hookenv.principal_unit()

    if unit:
        # Note(aluria): use Juju env var available since 2017

        return unit
    else:
        # Note(aluria): lookup all available IPv4/IPv6 addresses (except lo)
        ip_addresses = set()

        for iface in netifaces.interfaces():
            if iface == "lo":
                continue
            ip_addrs = netifaces.ifaddresses(iface)

            for iface_type in ip_addrs:
                if iface_type in (netifaces.AF_INET, netifaces.AF_INET6):
                    for addrs in ip_addrs[iface_type]:
                        ip_addresses.add(addrs["addr"])

        # Note(aluria): and try to match them against rel['private-address']

        for rel_type in hookenv.metadata()["requires"].keys():
            for rel in hookenv.relations_of_type(rel_type):
                if rel["private-address"] in ip_addresses:
                    return rel["__unit__"]


def get_disabled_plugins():
    """Return consolidated list of all plugins to be disabled."""
    config = hookenv.config()
    str_disabled_plugins = config["disabled_plugins"]
    disabled_plugins = str_disabled_plugins.split(":") if str_disabled_plugins else []
    # if LXD container, disable a list of unsuitable plugins
    if is_container():
        disabled_plugins_lxd = {
            "cpu",
            "disk",
            "diskio",
            "mem",
            "swap",
            "system",
            "bcache",
            "bond",
            "cgroup",
            "kernel_vmstat",
            "zoneinfo",
            "buddyinfo",
            "softnet_stat",
        }
        disabled_plugins_lxd.update(set(disabled_plugins))
        disabled_plugins = sorted(disabled_plugins_lxd)
    return disabled_plugins


def get_base_inputs():
    """Make a structure for rendering the base_inputs template.

    Returns a dict of items for the template.
    """
    extra_options = get_extra_options()
    conntrack = os.path.exists("/proc/sys/net/netfilter/nf_conntrack_max")
    config = hookenv.config()
    iptables = config["collect_iptables_metrics"]
    smart = config["collect_smart_metrics"]
    disabled_plugins = get_disabled_plugins()

    return {
        "extra_options": extra_options["inputs"],
        "bcache": is_bcache(),
        "python": sys.executable,
        "files_dir": get_files_dir(),
        "disabled_plugins": disabled_plugins,
        "conntrack": conntrack,
        "iptables": iptables,
        "smart": smart,
    }


def render_base_inputs():
    # use base inputs from charm templates
    with open(os.path.join(get_templates_dir(), "base_inputs.conf"), "r") as fd:
        return render_template(fd.read(), get_base_inputs())


def get_extra_options():
    extra_options = {"inputs": {}, "outputs": {}}
    extra_options_raw = hookenv.config()["extra_options"]
    extra_opts = yaml.full_load(extra_options_raw) or {}
    extra_options.update(extra_opts)
    # jsonify value, required as the telegraf config values format is similar
    # to raw json
    json_vals = {}
    # kind level

    for k, v in extra_options.items():
        json_vals[k] = {}
        # plugins level

        for plugin, values in v.items():
            json_vals[k][plugin] = {}
            # inner plugin (aka key:value)

            for key, val in values.items():
                if key in ("tagpass", "tagdrop"):
                    # this is a tagpass/drop, we need to go deeper
                    json_vals[k][plugin][key] = {}

                    for tag, tagvalue in val.items():
                        json_vals[k][plugin][key][tag] = json.dumps(tagvalue)
                else:
                    json_vals[k][plugin][key] = json.dumps(val)

    return json_vals


def render_extra_options(kind, name, extra_options=None):
    template = """
  {% if extra_options %}
  {% for key, value in extra_options.items() %}
  {% if key != 'tagpass' and key != 'tagdrop' %}
  {{ key }} = {{ value }}
  {% endif %}
  {% endfor %}
  {% for key, value in extra_options.items() %}
  {% if key == 'tagpass' or key == 'tagdrop' %}
  [{{ kind }}.{{ name }}.{{ key }}]
    {% for tag, tagvalue in value.items() %}
    {{ tag }} = {{ tagvalue }}
    {% endfor %}
  {% endif %}
  {% endfor %}
  {% endif %}
  """
    if extra_options is None:
        extra_options = get_extra_options()
    context = {
        "extra_options": extra_options[kind].get(name, {}),
        "kind": kind,
        "name": name,
    }
    return render_template(template, context)


def render_template(template, context):
    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)


def check_prometheus_port(key, new_port):
    unitdata_key = "{}.port".format(key)
    kv = unitdata.kv()
    existing_port = kv.get(unitdata_key)
    if existing_port != new_port:
        hookenv.open_port(new_port)
        hookenv.log("Opened port {}".format(new_port))
        if existing_port:  # Dont try to close non existing ports
            hookenv.close_port(existing_port)
            charms.promreg.deregister(None, existing_port)
        kv.set(unitdata_key, new_port)

    # Register the port, even if this is a re-registration
    charms.promreg.register(None, new_port)


def get_prometheus_port():
    config = hookenv.config()
    if not config.get("prometheus_output_port", False):
        return False
    if config.get("prometheus_output_port") == "default":
        return 9103
    return int(config.get("prometheus_output_port"))


def get_prometheus_ip_range():
    config = hookenv.config()
    if config.get("prometheus_ip_range") == "":
        return []
    # we should have a list of IPs, confirm that's the case
    ips = []
    for ip in config.get("prometheus_ip_range").split(","):
        # strip any spaces
        ip = ip.strip()
        try:
            ipaddress.ip_network(ip)
        except ValueError:
            hookenv.log(
                "Invalid prometheus_ip_range provided: {}".format(
                    config.get("prometheus_ip_range")
                ),
                level=hookenv.ERROR,
            )
            raise InvalidPrometheusIPRange()
        else:
            ips.append(ip)
    return ips


def get_socket_listener_port():
    config = hookenv.config()
    if not config.get("socket_listener_port", False):
        return 8094
    else:
        return int(config.get("socket_listener_port"))


def render_socket_listener_config(context):
    # Remove deprecated config file
    old_config_path = "{}/tcp_listener.conf".format(get_configs_dir())
    if os.path.exists(old_config_path):
        os.remove(old_config_path)

    config_path = "{}/socket_listener.conf".format(get_configs_dir())
    render(
        source="socket_listener.tmpl",
        templates_dir=get_templates_dir(),
        target=config_path,
        context=context,
    )


def get_sysstat_config_with_sadc_xall(content):
    """Get updated sysstat config content with `-S XALL` in `SADC_OPTIONS`.

    `/etc/sysstat/systat` consists of a sequence of shell variable assignments
    used to configure sysstat logging.

    Check the original config content.

    If change needed, make the change and return new config content.

    If no change, return None.
    """
    # if SADC_OPTIONS already exists with `-S XALL` in value, no need to change,
    # return None
    if re.search(r'^SADC_OPTIONS=".*-S\s+XALL.*"', content, flags=re.M):
        return None

    # if no such option, append and return
    # this also ignores commented option, e.g.: # SADC_OPTIONS="-S ALL"
    if not re.search(r'^SADC_OPTIONS=".*"', content, flags=re.M):
        return content.rstrip() + '\nSADC_OPTIONS="-S XALL"'

    match = re.search(r'^SADC_OPTIONS="(.*)"', content, flags=re.M)
    if match:
        # if option exists, check the value in quotes
        value = match.group(1)
        if "-S" in value:
            # if `-S` in value, replace it to XALL, keep other options unchanged.
            return re.sub(
                r'^SADC_OPTIONS="(.*-S\s+)(\w+)(.*)"',
                r'SADC_OPTIONS="\1XALL\3"',
                content,
                flags=re.M,
            )
        else:
            # if `-S` not in value, append `-S XALL` to value
            return re.sub(
                r'^SADC_OPTIONS="(.*)"',
                r'SADC_OPTIONS="\1 -S XALL"',
                content,
                flags=re.M,
            )


def update_sysstat_config_with_sdac_xall(path="/etc/sysstat/sysstat"):
    """Update `/etc/sysstat/sysstat` to ensure `-S XALL` in `SADC_OPTIONS`."""
    if os.path.isfile(path):
        with open(path, mode="r", encoding="utf8") as f:
            new_text = get_sysstat_config_with_sadc_xall(f.read())

        if new_text:
            hookenv.log("updating {} to ensure `-S XALL` in SADC_OPTIONS".format(path))
            with open(path, mode="w", encoding="utf8") as f:
                f.write(new_text)
    else:
        hookenv.log(
            "sysstat config file not found: {}".format(path), level=hookenv.WARNING
        )


def configure_telegraf():  # noqa: C901
    hookenv.log("Generating telegraf.conf", level=hookenv.DEBUG)
    update_sysstat_config_with_sdac_xall()
    config = hookenv.config()
    context = config.copy()
    try:
        config_path = get_main_config_path()
    except InvalidInstallMethod:
        hookenv.status_set(
            "blocked",
            "Wrong install_method provided: {!r}".format(config["install_method"]),
        )

        return

    if get_remote_unit_name() is None:
        hookenv.status_set("waiting", "Waiting for juju-info relation")
        # if UNIT_NAME in hostname config and relation not yet available,
        # make telegraf unable to start to not get weird metrics names

        if os.path.exists(config_path):
            os.unlink(config_path)

        return

    inputs = config.get("inputs_config", "")
    outputs = config.get("outputs_config", "")
    # just for the migration out of base64

    if inputs:
        try:
            inputs = base64.b64decode(inputs.encode("utf-8"), validate=True).decode(
                "utf-8"
            )
        except binascii.Error:
            # not bas64, probably already up to date configs
            pass

    if outputs:
        try:
            outputs = base64.b64decode(outputs.encode("utf-8"), validate=True).decode(
                "utf-8"
            )
        except binascii.Error:
            # not bas64, probably already up to date configs
            pass

    tags = []
    # Initialize juju-related tags to be exposed as labels
    # Using a dict to avoid duplicates, which telegraf will see as conflicts
    tags_dict = collections.OrderedDict()
    tags_dict["juju_application"] = get_remote_unit_name().split("/")[0]
    tags_dict["juju_unit"] = get_remote_unit_name().replace("/", "-")
    try:
        tags_dict["juju_model"] = hookenv.model_name()
    except KeyError:
        pass  # support older Juju 1.x deploys
    # Parse juju-configured tags, which can override the 3 juju_* defined above
    if config["tags"]:
        for tag in config["tags"].split(","):
            key, value = tag.split("=")
            tags_dict[key] = value
    # Render the tags dict into an array
    for key, value in tags_dict.items():
        tags.append('{} = "{}"'.format(key, value))
    context["tags"] = tags

    if inputs:
        context["inputs"] = inputs
    else:
        # use base inputs from charm templates
        context["inputs"] = render_base_inputs()

    if outputs:
        context["outputs"] = outputs
    else:
        context["outputs"] = ""
        hookenv.log("No output plugins in main config.")

    context["hostname"] = get_hostname_label()

    logfile_path = os.path.normpath(context["logfile"])

    if (
        context["logfile"]
        and not logfile_path.startswith("/var/log/")  # noqa W503
        and not (  # noqa W503
            config["install_method"] == "snap"
            and logfile_path.startswith("/var/snap/telegraf/common/")  # noqa W503
        )
    ):
        # only allow logging in /var/log, syslog, or /var/snap/telegraf/common
        hookenv.log(
            "logfile value reset to stderr."
            " Original value: {}".format(context["logfile"]),
            hookenv.DEBUG,
        )
        context["logfile"] = ""

    context["extra_options"] = get_extra_options()

    if get_socket_listener_port():
        context["socket_listener_port"] = get_socket_listener_port()

    render_socket_listener_config(context=context)

    disabled_plugins = ":".join(get_disabled_plugins())

    # handle plugin configuration for OVS
    if host.service_running("openvswitch-switch"):
        # add sudoers file for telegraf if openvswitch is running
        sudoers_filename = "telegraf_ovs"
        render_sudoers_file(sudoers_filename)
    else:
        # disable the OVS checks if the service is not curently running
        # no need to handle duplicates here, as those are handled during
        # template rendering later
        ovs_metric_names = [
            metric_name for metric_name in METRICS if metric_name.startswith("ovs_")
        ]

        hookenv.log(
            "disabling the following metrics, since OVS is not available: {}".format(
                ovs_metric_names
            )
        )

        ovs_disabled_plugins = ":".join(ovs_metric_names)

        if disabled_plugins:
            disabled_plugins = disabled_plugins + ":" + ovs_disabled_plugins
        else:
            disabled_plugins = ovs_disabled_plugins

    # handle the sudoers for iptables
    sudoers_filename = "telegraf_iptables"
    if config["collect_iptables_metrics"]:
        render_sudoers_file(sudoers_filename)
    else:
        remove_sudoers_file(sudoers_filename)

    # handle the sudoers for smart
    sudoers_filename = "telegraf_smart"
    if config["collect_smart_metrics"]:
        render_sudoers_file(sudoers_filename)
    else:
        remove_sudoers_file(sudoers_filename)

    telegraf_exec_metrics = os.path.join(get_files_dir(), "telegraf_exec_metrics.py")
    cmd = [
        telegraf_exec_metrics,
        "--render-config-files",
        "--disabled-metrics",
        disabled_plugins,
        "--configs-dir",
        get_configs_dir(),  # unit test can monkeypatch this path
        "--python",
        sys.executable,  # this will be venv python interpreter path
    ]
    hookenv.log("Rendering exec metrics config files: {}".format(" ".join(cmd)))
    subprocess.check_call(cmd)

    hookenv.log("Updating main config file")
    render(
        source="telegraf.conf.tmpl",
        templates_dir=get_templates_dir(),
        target=config_path,
        context=context,
    )

    # Make sure that only the right service is enabled, then defer to start_or_restart()
    for service in [DEB_SERVICE, SNAP_SERVICE]:
        if service == get_service():
            host.service_resume(service)
            host.service_reload(service)
        else:
            try:
                host.service_pause(service)
            except ValueError:
                # On machines not supporting snaps, it won't find the service
                # If we're using the snap, and the deb failed to install, it's fine too
                pass

    set_flag("telegraf.configured")
    set_flag("telegraf.needs_reload")

    if config["install_method"] == "deb":
        set_flag("telegraf.apt.configured")
    else:
        set_flag("telegraf.snap.configured")


# States


@when_not("telegraf.installed")
def install_telegraf():
    try:
        install_method = get_install_method()
    except InvalidInstallMethod:
        hookenv.status_set(
            "blocked", "Wrong install_method provided. Expected either 'deb' or 'snap'."
        )

        return

    if install_method == "deb":
        try:
            snap.remove("telegraf")
        except Exception:
            # the snap may already be absent, or snaps may not even be supported
            # in this environment
            pass
        apt.queue_install(["telegraf"])
    elif install_method == "snap":
        apt.purge("telegraf")
        config = hookenv.config()
        snap_channel = config.get("snap_channel")
        snap.install("telegraf", channel=snap_channel, classic=True)

    if install_method:
        set_flag("telegraf.installed")


@when("telegraf.installed")
@when("apt.installed.telegraf")
@when("plugins.prometheus-client.configured")
@when_not("telegraf.configured")
@when_not("telegraf.apt.configured")
def configure_telegraf_deb():
    configure_telegraf()


@when("telegraf.installed")
@when("snap.installed.telegraf")
@when("plugins.prometheus-client.configured")
@when_not("telegraf.configured")
@when_not("telegraf.snap.configured")
def configure_telegraf_snap():
    configure_telegraf()


@hook("upgrade-charm")
def upgrade_charm():
    for plugin in list_supported_plugins():
        clear_flag("plugins.{}.configured".format(plugin))
    clear_flag("extra_plugins.configured")
    clear_flag("telegraf.configured")
    clear_flag("telegraf.apt.configured")
    clear_flag("telegraf.snap.configured")
    clear_flag("grafana.configured")


@when("config.changed")
def handle_config_changes():
    config = hookenv.config()

    if config.changed("extra_options"):
        for plugin in list_supported_plugins():
            clear_flag("plugins.{}.configured".format(plugin))
        clear_flag("prometheus-client.relation.configured")
    # if something else changed, let's reconfigure telegraf itself just in case

    if config.changed("extra_plugins"):
        clear_flag("extra_plugins.configured")

    if config.get("collect_smart_metrics"):
        set_flag("telegraf.smart_metrics.enabled")
    else:
        clear_flag("telegraf.smart_metrics.enabled")

    if (
        config.changed("install_method")
        or config.changed("snap_channel")  # noqa W503
        or config.changed("install_sources")  # noqa W503
    ):
        clear_flag("telegraf.installed")
        clear_flag("extra_plugins.configured")
        clear_flag("plugins.prometheus-client.configured")
        clear_flag("prometheus-client.relation.configured")

    if config.changed("prometheus_output_port") or config.changed(
        "prometheus_ip_range"
    ):
        clear_flag("plugins.prometheus-client.configured")
        clear_flag("prometheus-client.relation.configured")
    clear_flag("telegraf.configured")
    clear_flag("telegraf.apt.configured")
    clear_flag("telegraf.snap.configured")
    clear_flag("telegraf.nagios-setup.complete")
    clear_flag("grafana.configured")


@when("telegraf.configured")
@when_not("extra_plugins.configured")
def configure_extra_plugins():
    config = hookenv.config()
    plugins = config["extra_plugins"]

    if plugins:
        config_path = "{}/extra_plugins.conf".format(get_configs_dir())
        host.write_file(config_path, plugins.encode("utf-8"))
        set_flag("extra_plugins.configured")
        set_flag("telegraf.needs_reload")


@when("elasticsearch.available")
@when("telegraf.installed")
def elasticsearch_input(es):
    template = """
[[inputs.elasticsearch]]
  servers = {{ servers }}
"""
    hosts = []
    rels = hookenv.relations_of_type("elasticsearch")
    for rel in rels:
        es_host = rel.get("host")
        port = rel.get("port")
        if not es_host or not port:
            hookenv.log("No host received for relation: {}.".format(rel))
            continue
        hosts.append("http://{}:{}".format(es_host, port))
    config_path = "{}/{}.conf".format(get_configs_dir(), "elasticsearch")
    if hosts:
        context = {"servers": json.dumps(hosts)}
        input_config = render_template(template, context) + render_extra_options(
            "inputs", "elasticsearch"
        )
        hookenv.log("Updating {} plugin config file".format("elasticsearch"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.elasticsearch.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
        clear_flag("plugins.elasticsearch.configured")
    set_flag("telegraf.needs_reload")


@when("memcached.available")
@when("telegraf.installed")
def memcached_input(memcache):
    template = """
[[inputs.memcached]]
  servers = {{ servers }}
"""
    required_keys = ["host", "port"]
    rels = hookenv.relations_of_type("memcached")
    addresses = []
    for rel in rels:
        if all([rel.get(key) for key in required_keys]):
            addr = rel["host"]
            port = rel["port"]
            address = "{}:{}".format(addr, port)
            addresses.append(address)
    config_path = "{}/{}.conf".format(get_configs_dir(), "memcached")
    if addresses:
        context = {"servers": json.dumps(addresses)}
        input_config = render_template(template, context) + render_extra_options(
            "inputs", "memcached"
        )
        hookenv.log("Updating {} plugin config file".format("memcached"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.memcached.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when("mongodb.database.available")
@when("telegraf.installed")
def mongodb_input(mongodb):
    template = """
[[inputs.mongodb]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type("mongodb")
    mongo_addresses = []
    for rel in rels:
        addr = rel["private-address"]
        port = rel.get("port", None)
        if port:
            mongo_address = "{}:{}".format(addr, port)
        else:
            mongo_address = addr
        mongo_addresses.append(mongo_address)
    config_path = "{}/{}.conf".format(get_configs_dir(), "mongodb")
    if mongo_addresses:
        context = {"servers": json.dumps(mongo_addresses)}
        input_config = render_template(template, context) + render_extra_options(
            "inputs", "mongodb"
        )
        hookenv.log("Updating {} plugin config file".format("mongodb"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.mongodb.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when("mysql.available")
@when("telegraf.installed")
def mysql_input(mysql):
    contexts = []
    for relid, relation in context.Relations()[mysql.relation_name].items():
        for unit, reldata in relation.items():
            hookenv.log("Available relations", level="DEBUG")
            if reldata["private-address"] == hookenv.unit_private_ip():
                if mysql.connection_string():
                    contexts.append(
                        {
                            "host": mysql.host(),
                            "port": mysql.port(),
                            "user": mysql.user(),
                            "pass": mysql.password(),
                            "is_secure": "false",  # TODO: provide config intf for this
                            "slave": reldata.get("slave", None),
                        }
                    )
                    break

    render_mysql_tmpl(contexts)
    toggle_flag("plugins.mysql.configured", bool(contexts))
    set_flag("telegraf.needs_reload")


@when("postgresql.database.connected")
def choose_postgresql_database():
    # We have no need for our own database. If we need to create
    # custom views, we can remove this handler. Or maybe it should
    # be configurable, so people can use telegraf to export custom
    # metrics about their data.
    pgsql = endpoint_from_flag("postgresql.database.connected")
    pgsql.set_database("postgres")


@when("postgresql.database.connected")
@when_not("postgresql.database.available")
def postgresql_waiting():
    hookenv.status_set("waiting", "Waiting for PostgreSQL relation")


@when("postgresql.database.available", "postgresql.database.changed")
@when("telegraf.installed")
def postgresql_input():
    # The subordinate may be connected to several services.
    contexts = []
    extra_options = get_extra_options()
    pg_options = extra_options["inputs"].get("postgresql")
    principal = hookenv.principal_unit()

    if principal not in hookenv.expected_related_units("postgresql"):
        return

    pgsql = endpoint_from_flag("postgresql.database.available")
    assert pgsql is not None
    for css in pgsql:
        cs = css.get(principal)
        if cs:
            ver = css.version or "9.3"
            contexts.append(
                {
                    "conn_str": str(cs),
                    "server": cs.host,
                    "replica": "master" if cs == css.master else "hot standby",
                    "extra_options": pg_options,
                    "version": ver,
                    "pg10": LooseVersion(ver + ".0") >= LooseVersion("10.0"),
                }
            )
            hookenv.status_set(
                "maintenance",
                "Monitoring PostgreSQL {} on {}".format(css.version, principal),
            )
    if not contexts:
        hookenv.status_set(
            "waiting", "Waiting for PostgreSQL database on {}".format(principal)
        )
        clear_flag("plugins.postgresql.configured")
        return

    render_postgresql_tmpl(contexts)
    set_flag("plugins.postgresql.configured")
    set_flag("telegraf.needs_reload")


def render_mysql_tmpl(contexts):
    config_path = "{}/{}.conf".format(get_configs_dir(), "mysql")

    if contexts:
        f = io.StringIO()
        template = open(os.path.join(get_templates_dir(), "mysql.tmpl"), "r").read()
        for ctx in contexts:
            f.write(render_template(template, ctx))
            f.write(render_extra_options("inputs", "mysql"))
        write_telegraf_file(config_path, f.getvalue())
    elif os.path.exists(config_path):
        os.unlink(config_path)


def render_postgresql_tmpl(contexts):
    config_path = "{}/{}.conf".format(get_configs_dir(), "postgresql")

    if contexts:
        f = io.StringIO()
        template = open(
            os.path.join(get_templates_dir(), "postgresql.tmpl"), "r"
        ).read()
        for ctx in contexts:
            f.write(render_template(template, ctx))
        write_telegraf_file(config_path, f.getvalue())
    elif os.path.exists(config_path):
        os.unlink(config_path)


@when("haproxy.available")
@when("telegraf.installed")
def haproxy_input(haproxy):
    template = """
[[inputs.haproxy]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type("haproxy")
    haproxy_addresses = []
    for rel in rels:
        enabled = rel.get("enabled", False)
        # Juju gives us a string instead of a boolean, fix it
        if isinstance(enabled, str):
            if enabled in ["y", "yes", "true", "t", "on", "True"]:
                enabled = True
            else:
                enabled = False
        if not enabled:
            continue
        addr = rel.get("listener-address", rel["private-address"])
        port = rel["port"]
        user = rel["user"]
        password = rel.get("password", None)
        userpass = user
        if password:
            userpass += ":{}".format(password)
        haproxy_address = "http://{}@{}:{}".format(userpass, addr, port)
        haproxy_addresses.append(haproxy_address)
    config_path = "{}/{}.conf".format(get_configs_dir(), "haproxy")
    if haproxy_addresses:
        input_config = render_template(
            template, {"servers": json.dumps(haproxy_addresses)}
        ) + render_extra_options("inputs", "haproxy")
        hookenv.log("Updating {} plugin config file".format("haproxy"))
        write_telegraf_file(config_path, input_config)
        set_flag("plugins.haproxy.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when("apache.available")
@when("telegraf.installed")
def apache_input(apache):
    template = """
[[inputs.apache]]
  urls = {{ urls }}
"""
    config_path = "{}/{}.conf".format(get_configs_dir(), "apache")
    port = "8080"
    vhost = render(
        source="apache-server-status.tmpl",
        templates_dir=get_templates_dir(),
        target=None,
        context={"port": port},
    )
    relation_info = {
        "ports": port,
        "domain": "apache-status",
        "enabled": True,
        "site_config": vhost,
        "site_modules": "status",
    }
    urls = []
    rels = hookenv.relations_of_type("apache")
    for rel in rels:
        hookenv.relation_set(rel["__relid__"], relation_settings=relation_info)
        addr = rel["private-address"]
        url = "http://{}:{}/server-status?auto".format(addr, port)
        urls.append(url)
    if urls:
        context = {"urls": json.dumps(urls)}
        input_config = render_template(template, context) + render_extra_options(
            "inputs", "apache"
        )
        hookenv.log("Updating {} plugin config file".format("apache"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.apache.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when("endpoint.redis.available")
@when("telegraf.installed")
def redis_input(redis):
    template = """
[[inputs.redis]]
  servers = ["tcp://{{ host }}:{{ port }}"]
  # Until https://github.com/influxdata/telegraf/issues/5036 is fixed
  fielddrop = ["aof_last_bgrewrite_status","aof_last_write_status","maxmemory_policy","rdb_last_bgsave_status","used_memory_dataset_perc","used_memory_peak_perc"]
"""  # noqa E501 (inline template)
    config_path = "{}/{}.conf".format(get_configs_dir(), "redis")

    rels = hookenv.relations_of_type("redis")
    if rels:
        if len(rels) != 1:
            hookenv.log(
                "Unexpected number of units in the redis relation."
                "Expected 1, got {}".format(len(rels)),
                "WARNING",
            )

        ctxt = {}
        ctxt["host"] = rels[0]["host"].strip('"')
        ctxt["port"] = rels[0]["port"]
        input_config = render_template(template, ctxt) + render_extra_options(
            "inputs", "redis"
        )
        hookenv.log("Updating {} plugin config file".format("redis"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.redis.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
        clear_flag("plugins.redis.configured")

    set_flag("telegraf.needs_reload")


@when("endpoint.sentry.joined")
@when("telegraf.installed")
def sentry_input(sentry):
    template = """
[[inputs.statsd]]
  protocol = "udp"
  service_address = ":8125"
  delete_gauges = false
  delete_counters = false
  delete_sets = false
  delete_timings = false
"""
    config_path = "{}/{}.conf".format(get_configs_dir(), "sentry")

    rels = hookenv.relations_of_type("sentry")
    if rels:
        for rel in rels:
            input_config = render_template(template, {}) + render_extra_options(
                "inputs", "sentry"
            )
            hookenv.log("Updating {} plugin config file".format("sentry"))
            host.write_file(config_path, input_config.encode("utf-8"))
            set_flag("plugins.sentry.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
        clear_flag("plugins.sentry.configured")

    set_flag("telegraf.needs_reload")


@when("exec.available")
@when("telegraf.installed")
def exec_input(exec_rel):
    template = """
{% for cmd in commands %}
[[inputs.exec]]
  commands = {{ cmd.commands }}
  {% for key, value in cmd|dictsort %}
      {% if key not in ["commands", "tags"] %}
  {{ key }} = "{{ value }}"
      {% endif %}
  {% endfor %}
  {% if cmd.tags %}
  [inputs.exec.tags]
    {% for tag, tag_value in cmd.tags|dictsort %}
    {{ tag }} = "{{ tag_value }}"
    {% endfor %}
  {% endif %}

{% endfor %}
"""
    config_path = "{}/{}.conf".format(get_configs_dir(), "exec")
    commands = exec_rel.commands()
    if not commands:
        hookenv.log("No Commands defined in the exec relation, doing nothing.")
        return
    pre_proc_cmds = []
    for command in commands:
        run_on_this_unit = command.pop("run_on_this_unit")
        if run_on_this_unit:
            pre_proc_cmds.append(command)
    if pre_proc_cmds:
        input_config = render_template(template, {"commands": pre_proc_cmds})
        hookenv.log("Updating {} plugin config file".format("exec"))
        host.write_file(config_path, input_config.encode("utf-8"))
        set_flag("plugins.exec.configured")
    else:
        # if no commands, remove previous config
        if os.path.exists(config_path):
            os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when_not("exec.available")
@when("plugins.exec.configured")
def exec_input_departed():
    config_path = "{}/{}.conf".format(get_configs_dir(), "exec")
    rels = hookenv.relations_of_type("exec")
    if not rels:
        clear_flag("plugins.exec.configured")
        if os.path.exists(config_path):
            os.unlink(config_path)
        set_flag("telegraf.needs_reload")


@when("amqp.connected")
@when_not("amqp.available")
def rabbitmq_input_setup(rabbitmq):
    # Requires management_plugin=true on the rabbitmq-server application.
    # vhost will not be used, but still needs to be requested.
    username = vhost = "telegraf-" + hookenv.local_unit().replace("/", "-")
    rabbitmq.set_local(admin=True)
    rabbitmq.set_remote(admin=True)
    rabbitmq.request_access(username=username, vhost=vhost)


@when("amqp.available")
@when("telegraf.installed")
def rabbitmq_input(rabbitmq):
    template = """
[[inputs.rabbitmq]]
  url = "http://{{ server }}:{{ port }}"
  username = "{{ username }}"
  password = "{{ password }}"
  fielddrop = ["idle_since"]
"""
    addr = rabbitmq.private_address()
    port = "15672"
    username = rabbitmq.username()
    password = rabbitmq.password()

    if not (addr and username and password):
        return

    config_path = "{}/{}.conf".format(get_configs_dir(), "rabbitmq")
    input_config = render_template(
        template,
        {"server": addr, "username": username, "password": password, "port": port},
    )

    hookenv.log("Updating {} plugin config file".format("rabbitmq"))
    write_telegraf_file(config_path, input_config)

    set_flag("plugins.rabbitmq.configured")
    set_flag("telegraf.needs_reload")


@when_not("amqp.available")
@when("plugins.rabbitmq.configured")
def rabbitmq_input_departed():
    config_path = "{}/{}.conf".format(get_configs_dir(), "rabbitmq")
    clear_flag("plugins.rabbitmq.configured")
    if os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


@when("influxdb-api.available")
@when("telegraf.installed")
def influxdb_api_output(influxdb):
    required_keys = ["hostname", "port", "user", "password"]
    rels = hookenv.relations_of_type("influxdb-api")
    endpoints = []
    user = None
    password = None
    for rel in rels:
        if all([rel.get(key) for key in required_keys]):
            endpoints.append("http://{}:{}".format(rel["hostname"], rel["port"]))
            if user is None:
                user = rel["user"]
            if password is None:
                password = rel["password"]
    config_path = "{}/{}.conf".format(get_configs_dir(), "influxdb-api")
    if endpoints:
        hookenv.log("Updating {} plugin config file".format("influxdb-api"))
        content = render(
            source="influxdb-api.conf.tmpl",
            target=None,
            templates_dir=get_templates_dir(),
            context={
                "urls": json.dumps(endpoints),
                "username": "{}".format(user),
                "password": "{}".format(password),
            },
        )
        extra_opts = render_extra_options("outputs", "influxdb")
        write_telegraf_file(config_path, "\n".join([content, extra_opts]))
        set_flag("plugins.influxdb-api.configured")
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_flag("telegraf.needs_reload")


def generate_prometheus_output_config(prometheus_output_port, prometheus_ip_range):
    # If extra_options are set for prometheus_client, let's integrate them
    extra_options = get_extra_options()
    options = extra_options["outputs"].get("prometheus_client", {})
    listen = options.pop("listen", None)
    if not listen:
        listen = ":{}".format(prometheus_output_port)
    elif int(listen.split(":", 1)[1]) != prometheus_output_port:
        hookenv.log(
            """prometheus_output_port is {}, but extra_options would set it
                       to {}. Choosing {} from prometheus_output_port.""".format(
                prometheus_output_port,
                int(listen.split(":", 1)[1]),
                prometheus_output_port,
            ),
            level=hookenv.WARNING,
        )
        listen = "{}:{}".format(listen.split(":", 1)[0], prometheus_output_port)

    options_ip_range = options.pop("ip_range", [])
    ip_range = options_ip_range + prometheus_ip_range

    return {
        "listen": listen,
        "ip_range": ip_range,
        "extra_options": render_extra_options(
            "outputs", "prometheus_client", extra_options=extra_options
        ),
    }


def render_prometheus_client_config(port, ip_range):
    config_path = "{}/{}.conf".format(get_configs_dir(), "prometheus_client")
    hookenv.log(
        "Updating {} plugin config file. Port is {} and ip_range is {}".format(
            "prometheus_client", port, ip_range
        ),
        level=hookenv.INFO,
    )
    context = generate_prometheus_output_config(port, ip_range)
    render(
        source="prometheus_client.tmpl",
        templates_dir=get_templates_dir(),
        target=config_path,
        context=context,
    )


@when("prometheus-client.available")
@when_not("prometheus-client.relation.configured")
def configure_prometheus_client_with_relation(prometheus):
    hookenv.log(
        "Configuring prometheus_client output plugin, with prometheus-client relation",
        level=hookenv.DEBUG,
    )
    port = get_prometheus_port() or "9103"
    # We'll iterate through the prometheus-client relation counterparts,
    # inform them of our address so that they scrape it, and get their egress subnets
    # so that we can allow them
    remote_egress_subnets = []
    for relation_id in hookenv.relation_ids("prometheus-client"):
        # if juju 2.x+ then we'll attempt to get the network space address
        try:
            hookenv.log("Getting local network info", level=hookenv.DEBUG)
            network_info = hookenv.network_get(
                "prometheus-client", relation_id=relation_id
            )
            hookenv.log(network_info, level=hookenv.DEBUG)
            if "ingress-addresses" in network_info:
                ip_addr = network_info.get("ingress-addresses")[0]
            else:
                ip_addr = hookenv.network_get_primary_address("prometheus-client")
            for unit in hookenv.related_units(relation_id):
                hookenv.log(
                    "Getting remote egress subnet for relation {} - {}".format(
                        unit, relation_id
                    ),
                    level=hookenv.DEBUG,
                )
                remote_egress_subnets.append(
                    hookenv.relation_get("egress-subnets", unit, relation_id)
                )
        except NotImplementedError:
            # if that fails, just let prometheus.configure(...) do it's default
            ip_addr = None
        prometheus.configure(port, hostname=ip_addr, private_address=ip_addr)
    check_prometheus_port("prometheus_output", port)
    # If prometheus_ip_range is empty, all remote IPs are allowed
    ip_range = get_prometheus_ip_range()
    if ip_range != []:
        ip_range = ip_range + remote_egress_subnets
    render_prometheus_client_config(port, ip_range)
    set_flag("plugins.prometheus-client.configured")
    set_flag("prometheus-client.relation.configured")
    set_flag("telegraf.needs_reload")


@when_not("prometheus-client.available")
@when_not("plugins.prometheus-client.configured")
def configure_prometheus_client():
    hookenv.log("Configuring prometheus_client output plugin", level=hookenv.DEBUG)
    if get_prometheus_port():
        port = get_prometheus_port()
    else:
        # No relation to prometheus, no port configured: do not configure the plugin
        set_flag("plugins.prometheus-client.configured")
        return
    check_prometheus_port("prometheus_output", port)
    ip_range = get_prometheus_ip_range()
    render_prometheus_client_config(port, ip_range)
    set_flag("plugins.prometheus-client.configured")
    set_flag("telegraf.needs_reload")
    clear_flag("prometheus-client.relation.configured")


def convert_days(time_string):
    """Convert string time descript to days.

    Function to convert strings like 2w or 14d to a sting containing the number
    of days.

    Not included, months and years, because the number of days in each changes.
    Also not included, seconds.
    """
    days = re.search(r"(\d+)d$", time_string)
    if days:
        return days.group(1)
    weeks = re.search(r"(\d+)w$", time_string)
    if weeks:
        days = int(weeks.group(1)) * 7
        return str(days)
    hours = re.search(r"(\d+)h$", time_string)
    if hours:
        days = int(hours.group(1)) / 24
        return str(days)
    mins = re.search(r"(\d+)m$", time_string)
    if mins:
        days = int(hours.group(1)) / 24 * 60
        return str(days)


@when("prometheus-rules.available")
def render_prometheus_rules(prometheus_rules):
    # Send a list of rules for alerting to Prometheus
    config = hookenv.config()
    unit_name = os.environ.get("JUJU_PRINCIPAL_UNIT")
    lead_time = config.get("lead_time")
    context = {
        "hostname": socket.gethostname(),
        "cpu_idle": config.get("cpu_idle"),
        "wait_time": config.get("wait_time"),
        "lead_time": lead_time,
        "lead_days": convert_days(lead_time),
        "prometheus_context": config.get("prometheus_context"),
        "unit_name": unit_name,
        "application_name": unit_name.split("/")[0],
    }
    formatted_rules = []
    template_files = [
        "rule_cpu_usage.j2",
        "rule_diskfull.j2",
        "rule_mem.j2",
        "rule_disk_ro.j2",
        "rule_packetdrops.j2",
        "rule_predict_disk_space.j2",
    ]
    for template_file in template_files:
        with open(os.path.join(get_templates_dir(), template_file), "r") as fd:
            formatted_rules.append(render_template(fd.read(), context))
    prometheus_rules.configure("\n".join(formatted_rules))


@when_not("prometheus-client.available")
@when("prometheus-client.relation.configured")
def prometheus_client_departed():
    hookenv.log("prometheus-client relation not available")
    config_path = "{}/{}.conf".format(get_configs_dir(), "prometheus_client")
    rels = hookenv.relations_of_type("prometheus-client")
    if not rels and os.path.exists(config_path):
        hookenv.log("Deleting {} plugin config file".format("prometheus-client"))
        os.unlink(config_path)
        clear_flag("plugins.prometheus-client.configured")


@when(
    "plugins.prometheus-client.configured",
    "endpoint.dashboards.joined",
    "leadership.is_leader",
)
@when_not("grafana.configured")
def register_grafana_dashboard():
    grafana = endpoint_from_flag("endpoint.dashboards.joined")
    hookenv.log("Loading grafana dashboard", level=hookenv.DEBUG)
    dashboard = _load_grafana_dashboard()
    digest = hashlib.md5(dashboard.encode("utf8")).hexdigest()
    dashboard_dict = json.loads(dashboard)
    dashboard_dict["digest"] = digest
    hookenv.log(
        "Rendered dashboard dict:\n{}".format(dashboard_dict), level=hookenv.DEBUG
    )
    grafana.register_dashboard(name=GRAFANA_DASHBOARD_NAME, dashboard=dashboard_dict)
    hookenv.log('Grafana dashboard "{}" registered.'.format(GRAFANA_DASHBOARD_NAME))
    set_flag("grafana.configured")


def _load_grafana_dashboard():
    prometheus_datasource = "{} - Juju generated source".format(
        hookenv.config().get("prometheus_datasource", "prometheus")
    )
    dashboard_context = dict(datasource=prometheus_datasource)
    # TODO: Figure out if metrics exist and then set bools accordingly.
    # For now, setting bools to true.
    dashboard_context["bonds_enabled"] = True
    dashboard_context["bcache_enabled"] = True
    dashboard_context["conntrack_enabled"] = True
    return render_custom(
        source=GRAFANA_DASHBOARD_TELEGRAF_FILE_NAME,
        render_context=dashboard_context,
        variable_start_string="<<",
        variable_end_string=">>",
    )


# This isn't exposed in charmhelpers: https://github.com/juju/charm-helpers/issues/367
def render_custom(source, render_context, **parameters):
    """Render a template from the template folder with custom environment parameters.

    source: template file name to render from
    context: template context variables
    parameters: initialization parameters for the jinja Environment

    returns the rendered template content
    """
    template_folder = os.path.join(hookenv.charm_dir(), "templates/dashboards/grafana")
    environment = Environment(loader=FileSystemLoader(template_folder), **parameters)
    try:
        template = environment.get_template(source)
    except exceptions.TemplateNotFound as e:
        hookenv.log(
            "Could not load template {} from {}".format(source, template_folder)
        )
        raise e
    return template.render(render_context)


@when("endpoint.dashboards.departed", "grafana.configured")
def unregister_grafana_dashboard():
    clear_flag("grafana.configured")


@when("endpoint.dashboards.failed", "leadership.is_leader")
def grafana_dashboard_import_failed():
    grafana = endpoint_from_flag("endpoint.dashboards.failed")
    for failed_import in grafana.failed_imports:
        hookenv.log(
            message='Grafana dashboard "{}" import failed with: {}'.format(
                failed_import.name, failed_import.reason
            ),
            level=hookenv.ERROR,
        )
    clear_flag("grafana.configured")


@when("telegraf.needs_reload")
@when("telegraf.installed")
@when("telegraf.configured")
def start_or_restart():
    states = sorted(
        [
            k
            for k in get_states().keys()
            if k.startswith("plugins") or k.startswith("extra_plugins")
        ]
    )

    service = get_service()
    config_files_changed = helpers.any_file_changed(list_config_files())
    active_plugins_changed = helpers.data_changed("active_plugins", states or "")
    if (
        not host.service_running(service)
        or config_files_changed  # noqa W503
        or active_plugins_changed  # noqa W503
    ):
        hookenv.log("Restarting telegraf")
        host.service_restart(service)
    else:
        hookenv.log(
            "Not restarting: active_plugins_changed={} | "
            "config_files_changed={}".format(
                active_plugins_changed, config_files_changed
            )
        )

    # Give telegraf time to restart.
    timeout = time.time() + 15
    while not host.service_running(service) and time.time() < timeout:
        time.sleep(0.1)

    if host.service_running(service):
        revision = ""
        if os.path.exists("version"):
            with open("version") as f:
                line = f.readline().strip()
            # We only want the first 8 characters, that's enough to tell
            # which version of the charm we're using.
            if len(line) > 8:
                revision = " (source version/commit {}…)".format(line[:8])
            else:
                revision = " (source version/commit {})".format(line)
        hookenv.status_set(
            "active", "Monitoring {}{}".format(get_remote_unit_name(), revision)
        )
        clear_flag("telegraf.needs_reload")
    else:
        hookenv.status_set("blocked", "Telegraf failed to start. Check config.")


def is_bcache():
    """Determine if this is a container.

    return true if bcache is present, and this is not a container.
    """
    container = is_container()
    return os.path.exists("/sys/fs/bcache") and not container


@hook("update-status")
def update_status():
    changed = charms.reactive.helpers.data_changed(
        "detect_changes", get_base_inputs(), hash_type="sha256"
    )
    if changed:
        clear_flag("telegraf.configured")
        clear_flag("telegraf.apt.configured")
        clear_flag("telegraf.snap.configured")


@when("nrpe-external-master.available")
@when("telegraf.installed")
@when("telegraf.configured")
@when_not("telegraf.nagios-setup.complete")
def configure_nagios(nagios):
    """Configure nagios process check.

    The flag 'telegraf.nagios-setup.complete' is reset at the moment config is
    changed, so this should make sure that updates are handled.
    """
    # Use charmhelpers to handle the configuration of nrpe
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=False)

    # use charmhelpers to create a process check
    nrpe_setup.add_check(
        "telegraf_http",
        "Telegraf HTTP check",
        "check_http -I 127.0.0.1 -p {} -u /metrics".format(get_prometheus_port()),
    )
    nrpe_setup.write()
    set_flag("telegraf.nagios-setup.complete")


@when("telegraf.smart_metrics.enabled")
@when_not("apt.smartmontools.installed")
@when_not("apt.nvme.installed")
def install_smart_metrics_packages():
    apt.queue_install(["smartmontools", "nvme-cli"])
