# Copyright 2015-2016 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import unittest
from unittest.mock import ANY, MagicMock, call, patch, sentinel

# Mock layer modules
import charms

apt = MagicMock()
layer = MagicMock()
promreg = MagicMock()
charms.apt = apt
charms.layer = layer
charms.promreg = promreg
sys.modules["charms.apt"] = apt
sys.modules["charms.layer"] = layer
sys.modules["charms.promreg"] = promreg

import reactive  # noqa E402
from reactive import telegraf  # noqa E402


class TestPostgreSQL(unittest.TestCase):
    @patch("reactive.telegraf.endpoint_from_flag")
    def test_choose_postgresql_database(self, endpoint_from_flag):
        telegraf.choose_postgresql_database()
        endpoint_from_flag.assert_called_once_with("postgresql.database.connected")
        endpoint_from_flag().set_database.assert_called_once_with("postgres")

    @patch("reactive.telegraf.hookenv.principal_unit")
    @patch("reactive.telegraf.hookenv.expected_related_units")
    @patch("reactive.telegraf.endpoint_from_flag")
    @patch("reactive.telegraf.get_extra_options")
    def test_postgresql_input_principal_not_postgresql(
        self,
        get_extra_options,
        endpoint_from_flag,
        expected_related_units,
        principal_unit,
    ):
        """Skip execution when the principal unit is not of type postgresql-charm."""
        principal_unit.return_value = sentinel.principal_unit
        expected_related_units.return_value = []
        reactive.telegraf.postgresql_input()
        endpoint_from_flag.assert_not_called()

    @patch("reactive.telegraf.set_flag")
    @patch("reactive.telegraf.hookenv.principal_unit")
    @patch("reactive.telegraf.hookenv.expected_related_units")
    @patch("reactive.telegraf.endpoint_from_flag")
    @patch("reactive.telegraf.get_extra_options")
    @patch("reactive.telegraf.render_postgresql_tmpl")
    def test_postgresql_input(
        self,
        render,
        get_extra_options,
        endpoint_from_flag,
        expected_related_units,
        principal_unit,
        set_flag,
    ):
        principal_unit.return_value = sentinel.principal_unit
        expected_related_units.return_value = [principal_unit.return_value]

        cs = MagicMock()  # ConnectionString from interface:pgsql
        cs.__str__.return_value = "libpq connection string"
        cs.host = sentinel.host

        css = MagicMock()  # ConnectionStrings from interface:pgsql
        css.version = "10"
        css.get.return_value = cs
        css.master = cs

        pgsql = MagicMock()
        pgsql.__iter__.return_value = [css]
        endpoint_from_flag.return_value = pgsql

        get_extra_options.return_value = {"inputs": {}}

        reactive.telegraf.postgresql_input()

        endpoint_from_flag.assert_called_once_with("postgresql.database.available")
        css.get.assert_called_once_with(sentinel.principal_unit)
        render.assert_called_once_with(
            [
                {
                    "replica": "master",
                    "conn_str": "libpq connection string",
                    "server": sentinel.host,
                    "extra_options": None,
                    "version": "10",
                    "pg10": True,
                }
            ]
        )
        set_flag.assert_has_calls(
            [call("plugins.postgresql.configured"), call("telegraf.needs_reload")]
        )

    @patch("reactive.telegraf.write_telegraf_file")
    @patch("reactive.telegraf.get_templates_dir")
    @patch("reactive.telegraf.get_base_dir")
    def test_render_postgresql_tmpl(
        self, get_base_dir, get_templates_dir, write_telegraf_file
    ):
        get_templates_dir.return_value = os.path.join(
            os.path.dirname(__file__), "../..", "templates"
        )
        get_base_dir.return_value = "/etc/telegraf"

        reactive.telegraf.render_postgresql_tmpl(
            [{"replica": "master", "conn_str": "my_conn_str", "server": "my_server"}]
        )

        write_telegraf_file.assert_called_once_with(
            "/etc/telegraf/telegraf.d/postgresql.conf",
            ANY,
        )

    @patch("reactive.telegraf.write_telegraf_file")
    @patch("reactive.telegraf.get_templates_dir")
    @patch("reactive.telegraf.get_base_dir")
    def test_render_postgresql_tmpl_extra_options(
        self, get_base_dir, get_templates_dir, write_telegraf_file
    ):
        get_templates_dir.return_value = os.path.join(
            os.path.dirname(__file__), "../..", "templates"
        )
        get_base_dir.return_value = "/etc/telegraf"

        pg_extra_opts = {
            "tagexclude": ["server"],
            "tagdrop": {"db": ["template*", "postgres", "telegraf"]},
        }

        context = {
            "replica": "master",
            "conn_str": "my_conn_str",
            "server": "my_server",
            "extra_options": pg_extra_opts,
        }
        reactive.telegraf.render_postgresql_tmpl([context])

        with open(
            os.path.join(telegraf.get_templates_dir(), "postgresql.tmpl"), "r"
        ) as fd:
            rendered = telegraf.render_template(fd.read(), context)

        write_telegraf_file.assert_called_once_with(
            "/etc/telegraf/telegraf.d/postgresql.conf", rendered
        )
        # now check if what's there is the thing we actually expect
        expected = """[[inputs.postgresql_extensible]]
  address = "my_conn_str"
  # Using the charm default server.
  outputaddress = "my_server"
  tagexclude = ['server']
  [inputs.postgresql_extensible.tagdrop]
    db = ['template*', 'postgres', 'telegraf']

  [inputs.postgresql_extensible.tags]
    replica = "master"

[[inputs.postgresql_extensible.query]]"""
        self.assertEqual(rendered.split("\n")[3:15], expected.split("\n"))

    @patch("os.unlink")
    @patch("os.path.exists")
    @patch("reactive.telegraf.get_base_dir")
    def test_render_postgresql_tmpl_no_context(self, get_base_dir, exists, unlink):
        exists.return_value = False
        get_base_dir.return_value = "/etc/telegraf"
        reactive.telegraf.render_postgresql_tmpl([])
        exists.assert_called_once_with("/etc/telegraf/telegraf.d/postgresql.conf")
        self.assertFalse(unlink.called)
        exists.return_value = True
        reactive.telegraf.render_postgresql_tmpl([])
        unlink.assert_called_once_with("/etc/telegraf/telegraf.d/postgresql.conf")
