# Copyright 2015-2016 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for actions.py."""
import base64
import getpass
import grp
import json
import os
import shutil
import subprocess
import sys
from textwrap import dedent
from unittest import mock
from unittest.mock import MagicMock, call, patch

from charmhelpers.core import host
from charmhelpers.core.hookenv import Config
from charmhelpers.core.templating import render

import charms
from charms.reactive import RelationBase, bus, helpers, set_flag

import py

import pytest

import yaml

# Mock layer modules
apt = MagicMock()
layer = MagicMock()
promreg = MagicMock()
charms.apt = apt
charms.layer = layer
charms.promreg = promreg
sys.modules["charms.apt"] = apt
sys.modules["charms.layer"] = layer
sys.modules["charms.promreg"] = promreg

import reactive  # noqa E402
from files.telegraf_exec_metrics import METRICS  # noqa E402
from reactive import telegraf  # noqa E402

UNIT_TESTS_DIR = os.path.dirname(os.path.abspath(__file__))
UNIT_TESTS_DATA_DIR = os.path.join(UNIT_TESTS_DIR, "data")


@pytest.fixture(autouse=True)
def setup(monkeypatch, tmpdir):
    monkeypatch.setitem(os.environ, "JUJU_UNIT_NAME", "telegraf/0")
    monkeypatch.setitem(os.environ, "JUJU_MODEL_NAME", "telegraf-test-model")
    monkeypatch.setattr(telegraf, "get_remote_unit_name", lambda: "remote-unit/0")
    # mock this to avoid permission errors
    monkeypatch.setattr(telegraf, "update_sysstat_config_with_sdac_xall", lambda: True)
    monkeypatch.setattr(telegraf.hookenv, "status_set", lambda status, msg: None)
    monkeypatch.setattr(telegraf.hookenv, "log", lambda msg, **kw: None)
    monkeypatch.setattr(telegraf.host, "service_resume", lambda svc: None)
    monkeypatch.setattr(telegraf.host, "service_reload", lambda svc: None)
    monkeypatch.setattr(telegraf.host, "service_pause", lambda svc: None)

    # patch host.write for non-root
    user = getpass.getuser()
    group = grp.getgrgid(os.getegid()).gr_name
    orig_write_file = telegraf.host.write_file

    def intercept_write_file(*a, **kw):
        # fix the args to use non-root, this is for callers that pass
        # owner/group as positional arguments like
        # charmhelpers.core.templating.render

        if len(a) > 2:
            if a[2] == "root" and a[3] == "root":
                # make all files writable by owner, as we need don't run as root
                a = (a[0], a[1], user, group, 0o744)
        else:
            kw["owner"] = user
            kw["group"] = group
            # make all files writable by owner, as we need don't run as root
            kw["perms"] = 0o744

        return orig_write_file(*a, **kw)

    monkeypatch.setattr(telegraf.host, "write_file", intercept_write_file)


@pytest.fixture(autouse=True)
def cleanup(request):
    def unit_state_cleanup():
        # cleanup unitdata
        from charmhelpers.core import unitdata

        unitdata._KV = None
        # rm unit-state.db file
        unit_state_db = os.path.join(telegraf.hookenv.charm_dir(), ".unit-state.db")

        if os.path.exists(unit_state_db):
            os.unlink(unit_state_db)

    request.addfinalizer(unit_state_cleanup)


@pytest.fixture()
def temp_charm_dir(monkeypatch, tmpdir):
    charm_dir = tmpdir.mkdir("charm_dir")
    os.environ["CHARM_DIR"] = charm_dir.strpath
    # also monkeypatch get_templates_dir to fix the path
    real_charm_dir = os.path.join(os.path.dirname(reactive.__file__), "../")
    monkeypatch.setattr(
        telegraf, "get_templates_dir", lambda: os.path.join(real_charm_dir, "templates")
    )
    monkeypatch.setattr(
        telegraf, "get_files_dir", lambda: os.path.join(real_charm_dir, "files")
    )
    # fix hookenv.metadata
    with open(os.path.join(real_charm_dir, "metadata.yaml")) as md:
        metadata = yaml.full_load(md)
    monkeypatch.setattr(telegraf.hookenv, "metadata", lambda: metadata)


@pytest.fixture(autouse=True)
def temp_config_dir(monkeypatch, tmpdir):
    base_dir = tmpdir.mkdir("etc_telegraf")
    base_dir.mkdir(telegraf.CONFIG_DIR)
    monkeypatch.setattr(telegraf, "DEB_BASE_DIR", base_dir.strpath)

    sudoers_dir = tmpdir.mkdir("etc_sudoers")
    monkeypatch.setattr(telegraf, "SUDOERS_DIR", sudoers_dir.strpath)


# For now, too many tests fails if we set also "snap" method
# TODO: fix all tests so every single one is called once with each install_method
@pytest.fixture(params=["deb"])
def install_method(request):
    return request.param


@pytest.fixture(autouse=True)
def config(monkeypatch, temp_charm_dir, install_method):
    raw_config = yaml.full_load(open("config.yaml", "r"))
    data = dict((k, v["default"]) for k, v in raw_config["options"].items())
    data["install_method"] = install_method
    config = Config(data)
    monkeypatch.setattr(telegraf.hookenv, "config", lambda: config)

    return config


# utilities
def base_dir():
    return py.path.local(telegraf.DEB_BASE_DIR)


def configs_dir():
    return py.path.local(telegraf.get_configs_dir())


def persist_state():
    """Fake persistent state by calling helpers that modify unitdata.kv."""
    states = [
        k
        for k in bus.get_states().keys()
        if k.startswith("plugins") or k.startswith("extra_plugins")
    ]
    helpers.any_file_changed(telegraf.list_config_files())

    if states:
        helpers.data_changed("active_plugins", states)


def run_telegraf_exec_metrics(*args):
    script = os.path.join(telegraf.get_files_dir(), "telegraf_exec_metrics.py")
    cmd = [script] + list(args)
    assert 0 == subprocess.call(cmd)


# Tests


def check_sysstat_config(original, expected):
    original = original.strip()
    expected = expected.strip()
    actual = telegraf.get_sysstat_config_with_sadc_xall(original)
    assert actual == expected


@pytest.mark.parametrize("install_method", ["deb", "snap"])
def test_write_telegraf_file(mocker, install_method):

    if install_method == "deb":
        user = telegraf.DEB_OWNER
        group = telegraf.DEB_GROUP
    elif install_method == "snap":
        user = telegraf.SNAP_OWNER
        group = telegraf.SNAP_GROUP
    write_file = mocker.patch("charmhelpers.core.host.write_file")
    telegraf.write_telegraf_file("the-path", "the-content áéíóú")
    write_file.assert_called_once_with(
        "the-path",
        "the-content áéíóú".encode("utf-8"),
        owner=user,
        group=group,
        perms=0o640,
    )


def test_sadc_options_correct(monkeypatch):
    """If SADC_OPTIONS is already correct, should return None."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-S XALL"
        # another comment
        UMASK=0022
    """
    )
    assert telegraf.get_sysstat_config_with_sadc_xall(original) is None


def test_sadc_options_correct_included(monkeypatch):
    """If SADC_OPTIONS already includes `-S XALL`, should return None."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-B before -S XALL -A after"
        # another comment
        UMASK=0022
    """
    )
    assert telegraf.get_sysstat_config_with_sadc_xall(original) is None


def test_sadc_options_non_exist(monkeypatch):
    """If SADC_OPTIONS doesn't exist, should just append."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-S XALL"
    """
    )

    check_sysstat_config(original, expected)


def test_sadc_options_commented(monkeypatch):
    """If SADC_OPTIONS exists but commented, should ignore and append."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        #SADC_OPTIONS="-S XALL"
        # another comment
        UMASK=0022
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        #SADC_OPTIONS="-S XALL"
        # another comment
        UMASK=0022
        SADC_OPTIONS="-S XALL"
    """
    )

    check_sysstat_config(original, expected)


def test_sadc_options_incorrect(monkeypatch):
    """If SADC_OPTIONS exists but not XALL, should replace."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-S DISK"
        # another comment
        UMASK=0022
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-S XALL"
        # another comment
        UMASK=0022
    """
    )
    check_sysstat_config(original, expected)


def test_sadc_options_incorrect_included(monkeypatch):
    """If SADC_OPTIONS exists but not XALL, should replace."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-B before -S DISK -A after"
        # another comment
        UMASK=0022
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-B before -S XALL -A after"
        # another comment
        UMASK=0022
    """
    )
    check_sysstat_config(original, expected)


def test_sadc_options_commented_line_not_touched(monkeypatch):
    """If SADC_OPTIONS exists but has no -S, should append."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        #SADC_OPTIONS="-C a-comment -D -V --arg1 --arg2"
        SADC_OPTIONS="-C a-comment -D -V"
        # another comment
        UMASK=0022
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        #SADC_OPTIONS="-C a-comment -D -V --arg1 --arg2"
        SADC_OPTIONS="-C a-comment -D -V -S XALL"
        # another comment
        UMASK=0022
    """
    )

    check_sysstat_config(original, expected)


def test_sadc_options_s_missing(monkeypatch):
    """If SADC_OPTIONS exists but has no -S, should append."""
    original = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-C a-comment -D -V"
        # another comment
        UMASK=0022
    """
    )

    expected = dedent(
        """
        # some comment
        COMPRESSAFTER=10
        SADC_OPTIONS="-C a-comment -D -V -S XALL"
        # another comment
        UMASK=0022
    """
    )

    check_sysstat_config(original, expected)


def test_telegraf_exec_metrics(monkeypatch, temp_config_dir):
    metrics = set(
        ["sockstat", "sockstat6", "softnet_stat", "buddyinfo", "zoneinfo", "netns"]
    )

    for metric in metrics:
        run_telegraf_exec_metrics("--metric", metric)

    # another test for netns, read cmd output from file
    run_telegraf_exec_metrics(
        "--metric",
        "netns",
        "--input-file",
        os.path.join(UNIT_TESTS_DATA_DIR, "netns.txt"),
    )
    # parse output form `sudo ovs-appctl dpctl/show -s`
    run_telegraf_exec_metrics(
        "--metric",
        "ovs_dpctl",
        "--input-file",
        os.path.join(UNIT_TESTS_DATA_DIR, "ovs_dpctl.txt"),
    )

    # ensure config files exists after render
    run_telegraf_exec_metrics("--render-config-files", "--configs-dir", configs_dir())

    for metric in metrics:
        assert os.path.isfile(os.path.join(configs_dir(), "{}.conf".format(metric)))

    # ensure --disabled-metrics option is working
    disabled_metrics = set(["sockstat", "zoneinfo"])
    run_telegraf_exec_metrics(
        "--render-config-files",
        "--configs-dir",
        configs_dir(),
        "--disabled-metrics",
        ":".join(disabled_metrics),
    )
    # config files for disabled metrics should be removed.

    for metric in disabled_metrics:
        assert not os.path.isfile(os.path.join(configs_dir(), "{}.conf".format(metric)))
    # config files for other metrics should still be there

    for metric in metrics - disabled_metrics:
        assert os.path.isfile(os.path.join(configs_dir(), "{}.conf".format(metric)))


def test_telegraf_exec_metrics_enable_ovs(monkeypatch, config, temp_config_dir):
    config["disabled_plugins"] = ""  # ensure no metrics are initially disabled
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    monkeypatch.setattr(host, "service_running", lambda svc: True)
    telegraf.configure_telegraf()  # this will generate the check definitions

    # get the names of all the OVS metrics
    ovs_metric_names = [
        metric_name for metric_name in METRICS if metric_name.startswith("ovs_")
    ]

    # check that the check definitions for the OVS checks are in place
    for metric in ovs_metric_names:
        assert os.path.isfile(os.path.join(configs_dir(), "{}.conf".format(metric)))


def test_telegraf_exec_metrics_disable_ovs(monkeypatch, config):
    def mock_check_call(*args, **kwargs):
        disabled_metrics = args[0][3]
        assert disabled_metrics  # this should not be empty
        for metric in disabled_metrics.split(":"):
            assert metric.startswith("ovs_")

    config["disabled_plugins"] = ""  # ensure no metrics are initially disabled
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    monkeypatch.setattr(host, "service_running", lambda svc: False)
    monkeypatch.setattr(subprocess, "check_call", mock_check_call)
    telegraf.configure_telegraf()


def test_ovs_dump_flows_cmd_metric(monkeypatch):
    def mock_check_output(*args, **kwargs):
        """Mock the subprocess calls to the OVS tools."""
        cmd = args[0][1]
        if cmd == "ovs-vsctl" and args[0][2] == "list-br":
            return b"br-ex\nbr-data\n"
        elif cmd == "ovs-vsctl" and args[0][6] == "name=br-ex":
            return (
                b'{"data":[[["uuid","8267c260-81d5-4a06-848f-711b82d9bf4b"],'
                b'["set",[]],["set",[]],"000062c26782064a","system","<unknown>",'
                b'["map",[]],["set",[]],["set",[]],["map",[]],["set",[]],false,'
                b'["set",[]],"br-ex",["set",[]],["map",[]],'
                b'["uuid","c1b763d1-a706-46af-98a8-4b4bd0125249"],["set",[]],false,'
                b'["map",[]],["set",[]],["map",[]],false]],'
                b'"headings":["_uuid","auto_attach","controller","datapath_id",'
                b'"datapath_type","datapath_version","external_ids","fail_mode",'
                b'"flood_vlans","flow_tables","ipfix","mcast_snooping_enable",'
                b'"mirrors","name","netflow","other_config","ports","protocols",'
                b'"rstp_enable","rstp_status","sflow","status","stp_enable"]}'
            )
        elif cmd == "ovs-vsctl" and args[0][6] == "name=br-data":
            return (
                b'{"data":[[["uuid","8267c260-81d5-4a06-848f-711b82d9bf4c"],'
                b'["set",[]],["set",[]],"000062c26782064b","system","<unknown>",'
                b'["map",[]],["set",[]],["set",[]],["map",[]],["set",[]],false,'
                b'["set",[]],"br-data",["set",[]],["map",[]],'
                b'["uuid","c1b763d1-a706-46af-98a8-4b4bd0125250"],'
                b'["set",["OpenFlow13", "OpenFlow15"]],false,["map",[]],'
                b'["set",[]],["map",[]],false]],'
                b'"headings":["_uuid","auto_attach","controller","datapath_id",'
                b'"datapath_type","datapath_version","external_ids","fail_mode",'
                b'"flood_vlans","flow_tables","ipfix","mcast_snooping_enable",'
                b'"mirrors","name","netflow","other_config","ports","protocols",'
                b'"rstp_enable","rstp_status","sflow","status","stp_enable"]}'
            )
        elif (
            cmd == "ovs-ofctl" and args[0][3] == "br-ex"
        ):  # sudo ovs-ofctl dump-flows br-ex
            return (
                b"NXST_FLOW reply (xid=0x4):\n"
                b" cookie=0x3711311ec90146a9, duration=247966.1s, table=0, n_packets=0,"
                b'n_bytes=0, priority=2,in_port="phy-br-ex" actions=drop\n'
                b" cookie=0x3711311ec90146a9, duration=247966.5s, table=0, n_packets=2,"
                b"n_bytes=156148, priority=0 actions=NORMAL\n"
            )
        elif (
            cmd == "ovs-ofctl" and args[0][5] == "br-data"
        ):  # sudo ovs-ofctl dump-flows br-data
            return (
                b"NXST_FLOW reply (xid=0x4):\n"
                b" cookie=0x3711311ec90146a0, duration=247966.1s, table=0, n_packets=0,"
                b'n_bytes=0, priority=2,in_port="phy-br-data" actions=drop\n'
                b" cookie=0x3711311ec90146a0, duration=247966.5s, table=0, n_packets=2,"
                b"n_bytes=156148, priority=0 actions=NORMAL\n"
            )
        else:
            raise Exception("unexpected argument: {}".format(args))

    monkeypatch.setattr(subprocess, "check_output", mock_check_output)
    monkeypatch.setattr(shutil, "which", lambda path: True)

    from files.telegraf_exec_metrics import OvsDumpFlowsCmdMetric

    metric = OvsDumpFlowsCmdMetric()

    raw_output = metric.get_input_content()
    parsed_output = metric.parse(raw_output)
    expected_output = [
        {"bridge": "br-ex", "count": 2},
        {"bridge": "br-data", "count": 2},
    ]
    assert parsed_output == expected_output


def test_get_remote_unit_name_juju2(monkeypatch):
    monkeypatch.undo()
    monkeypatch.setitem(os.environ, "JUJU_PRINCIPAL_UNIT", "remote/0")
    assert telegraf.get_remote_unit_name() == "remote/0"


def test_get_remote_unit_name_juju1(monkeypatch):
    monkeypatch.undo()
    monkeypatch.setattr(telegraf.netifaces, "interfaces", lambda: ["lo", "eth0"])

    ifaddresses = {
        17: [{"addr": "AA:BB:CC:DD:EE:FF", "broadcast": "ff:ff:ff:ff:ff:ff"}],
        2: [{"addr": "1.2.3.4", "netmask": "255.255.255.0", "broadcast": "1.2.3.1"}],
        10: [{"addr": "2620:0:862:ed1a::1", "netmask": "ffff:ffff:ffff:ffff::"}],
    }
    monkeypatch.setattr(
        telegraf.netifaces, "ifaddresses", lambda n: ifaddresses if n == "eth0" else {}
    )

    md = {
        "name": "telegraf",
        "requires": {
            "mysql": {"interface": "mysql-root"},
            "postgresql": {"interface": "pgsql"},
            "mongodb": {"interface": "mongodb", "scope": "container"},
            "memcached": {"interface": "memcache", "scope": "container"},
            "elasticsearch": {"interface": "elasticsearch", "scope": "container"},
            "haproxy": {"interface": "statistics", "scope": "container"},
            "apache": {"interface": "apache-website", "scope": "container"},
            "influxdb-api": {"interface": "influxdb-api"},
            "juju-info": {"interface": "juju-info", "scope": "container"},
            "exec": {"interface": "telegraf-exec", "scope": "container"},
        },
        "provides": {
            "prometheus-client": {"interface": "http"},
            "prometheus-rules": {"interface": "prometheus-rules"},
        },
    }
    monkeypatch.setattr(telegraf.hookenv, "metadata", lambda: md)

    def relation_ids(reltype):
        if reltype == "juju-info":
            return ["juju-info:0"]
        else:
            return []

    monkeypatch.setattr(telegraf.hookenv, "relation_ids", relation_ids)

    def related_units(relid):
        if relid == "juju-info:0":
            return ["remote/0"]
        else:
            return []

    monkeypatch.setattr(telegraf.hookenv, "related_units", related_units)

    relations = []
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    assert telegraf.get_remote_unit_name() is None

    relations = [{"private-address": "1.2.3.4", "__unit__": "remote/0"}]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    assert telegraf.get_remote_unit_name() == "remote/0"


def test_inputs_config_set(monkeypatch, config):
    config[
        "inputs_config"
    ] = """
    [[inputs.cpu]]
        percpu = true
"""

    def check(*a, **kw):
        assert kw["context"]["inputs"] == config["inputs_config"]

    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    monkeypatch.setattr(telegraf, "render", check)
    telegraf.configure_telegraf()


def test_old_base64_inputs_and_outputs(monkeypatch, config):
    config["inputs_config"] = base64.b64encode(
        b"""
    [[inputs.cpu]]
        percpu = true
"""
    ).decode("utf-8")
    config["outputs_config"] = base64.b64encode(
        b"""
    [[outputs.fake]]
        foo = true
"""
    ).decode("utf-8")
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)

    def check(*a, **kw):
        expected = base64.b64decode(config["inputs_config"]).decode("utf-8")
        assert kw["context"]["inputs"] == expected
        expected = base64.b64decode(config["outputs_config"]).decode("utf-8")
        assert kw["context"]["outputs"] == expected

    monkeypatch.setattr(telegraf, "render", check)
    telegraf.configure_telegraf()


def test_inputs_config_not_set(monkeypatch, config):
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    config["inputs_config"] = ""

    def check(*a, **kw):
        assert kw["context"]["inputs"] == telegraf.render_base_inputs()

    monkeypatch.setattr(telegraf, "render", check)
    telegraf.configure_telegraf()


def test_outputs_config(monkeypatch, config):
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    config[
        "outputs_config"
    ] = """
    [[outputs.foo]]
        server = "http://localhost:42"
"""

    def check(*a, **kw):
        assert kw["context"]["outputs"] == config["outputs_config"]

    monkeypatch.setattr(telegraf, "render", check)
    telegraf.configure_telegraf()


def test_extra_plugins(config):
    config[
        "extra_plugins"
    ] = """[[inputs.foo]]
    some_option = "http://foo.bar.com"
[[outputs.baz]]
    option = "enabled"
    """
    telegraf.configure_extra_plugins()
    assert configs_dir().join("extra_plugins.conf").read() == config["extra_plugins"]


def test_render_extra_options(config):
    extra_options = """
    inputs:
        test:
            boolean: true
            string: 10s
            list: ["a", "b"]
"""
    config["extra_options"] = extra_options
    content = telegraf.render_extra_options("inputs", "test")
    expected = """  boolean = true\n  list = ["a", "b"]\n  string = "10s"\n"""
    assert sorted(content.split()) == sorted(expected.split())


def test_get_extra_options(config):
    extra_options = """
    inputs:
        test:
            boolean: true
            string: somestring
            list: ["a", "b"]
            tagdrop:
                tag: ["foo", "bar"]
"""
    config["extra_options"] = extra_options
    extra_opts = telegraf.get_extra_options()
    expected = {
        "inputs": {
            "test": {
                "boolean": "true",
                "string": '"somestring"',
                "list": '["a", "b"]',
                "tagdrop": {"tag": '["foo", "bar"]'},
            }
        },
        "outputs": {},
    }
    assert extra_opts == expected


def test_render_extra_options_override(config):
    extra_options = """
    inputs:
        test:
            boolean: true
            string: 10s
            list: ["a", "b"]
"""
    config["extra_options"] = extra_options
    # clone extra_options and use a modified version
    options = {"inputs": {"test": {"string": json.dumps("20s")}}}
    content = telegraf.render_extra_options("inputs", "test", extra_options=options)
    expected = """  string = "20s"\n"""
    assert sorted(content.split()) == sorted(expected.split())


def test_render_base_inputs(config):
    base_inputs_opts = """
inputs:
    cpu:
        foo: 10s
        percpu: false
        fielddrop: ["time_*"]
        tagpass:
            cpu: ["cpu0"]
"""
    config["extra_options"] = base_inputs_opts
    content = telegraf.render_base_inputs()
    expected = """
# Read metrics about cpu usage
[[inputs.cpu]]
  fielddrop = ["time_*"]
  foo = "10s"
  percpu = false
  [inputs.cpu.tagpass]
    cpu = ["cpu0"]
"""
    assert content[: len(expected)] == expected


def test_base_inputs_disabled_plugins(config):
    """Check disabled_plugins option is working for builtin inputs."""
    config["disabled_plugins"] = "cpu:disk"
    content = telegraf.render_base_inputs()

    # disabled, should not exist
    assert "[[inputs.cpu]]" not in content
    assert "[[inputs.disk]]" not in content

    # not disabled, should exist
    assert "[[inputs.mem]]" in content


def test_base_inputs_disabled_lxd(monkeypatch, config):
    """Check that on a LXD container, builtin inputs are disabled."""
    config["disabled_plugins"] = "cpu:disk"
    monkeypatch.setattr(telegraf, "is_container", lambda: True)
    content = telegraf.render_base_inputs()
    # disabled, should not exist
    assert "[[inputs.cpu]]" not in content
    assert "[[inputs.disk]]" not in content
    assert "[[inputs.mem]]" not in content


def test_check_prometheus_port(monkeypatch):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    telegraf.check_prometheus_port("test_check_port", 10042)
    assert 10042 in open_ports


def test_check_prometheus_port_replace_old_port(monkeypatch):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    monkeypatch.setattr(telegraf.hookenv, "close_port", lambda p: open_ports.remove(p))
    telegraf.check_prometheus_port("test_check_port", 10042)
    assert 10042 in open_ports
    telegraf.check_prometheus_port("test_check_port", 10043)
    assert 10043 in open_ports
    assert 10042 not in open_ports


def test_get_prometheus_port(monkeypatch, config):
    config["prometheus_output_port"] = ""
    assert telegraf.get_prometheus_port() is False
    config["prometheus_output_port"] = "default"
    assert telegraf.get_prometheus_port() == 9103
    config["prometheus_output_port"] = "9126"
    assert telegraf.get_prometheus_port() == 9126


def test_prometheus_global(monkeypatch, config):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    monkeypatch.setattr(telegraf.hookenv, "close_port", lambda p: open_ports.remove(p))
    config["prometheus_output_port"] = "default"
    telegraf.configure_prometheus_client()
    expected = """[[outputs.prometheus_client]]
  listen = ":9103"
  ip_range = []"""
    config_file = base_dir().join("telegraf.d", "prometheus_client.conf")
    assert expected in config_file.read()


def test_prometheus_global_with_extra_options(monkeypatch, config):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    monkeypatch.setattr(telegraf.hookenv, "close_port", lambda p: open_ports.remove(p))
    config["prometheus_output_port"] = "default"
    config[
        "extra_options"
    ] = """
outputs:
  prometheus_client:
    namedrop: ["aerospike*"]
    tagpass:
      cpu: ["cpu0"]

"""
    telegraf.configure_prometheus_client()
    expected = """[[outputs.prometheus_client]]
  listen = ":9103"
  ip_range = []
  namedrop = ["aerospike*"]
  [outputs.prometheus_client.tagpass]
    cpu = ["cpu0"]"""
    config_file = base_dir().join("telegraf.d", "prometheus_client.conf")
    assert expected in config_file.read()


def test_default_tags_juju2(monkeypatch, config):
    monkeypatch.setattr(telegraf.hookenv, "principal_unit", lambda: "principal-unit")
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: None)
    config["tags"] = "juju_model=custom-model-name,juju_application=custom-app-name"
    telegraf.configure_telegraf()
    expected = """
[tags]
  # dc = "us-east-1" # will tag all metrics with dc=us-east-1
  # rack = "1a"

  juju_application = "custom-app-name"

  juju_unit = "remote-unit-0"

  juju_model = "custom-model-name"

"""
    config_file = base_dir().join("telegraf.conf")
    assert expected in config_file.read()


def test_default_tags_juju1(monkeypatch, config):
    monkeypatch.setattr(telegraf.hookenv, "principal_unit", lambda: "principal-unit")
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: None)
    monkeypatch.delitem(os.environ, "JUJU_MODEL_NAME")  # no models in Juju 1.x
    telegraf.configure_telegraf()
    expected = """
[tags]
  # dc = "us-east-1" # will tag all metrics with dc=us-east-1
  # rack = "1a"

  juju_application = "remote-unit"

  juju_unit = "remote-unit-0"

"""
    config_file = base_dir().join("telegraf.conf")
    assert expected in config_file.read()


# Plugin tests


def test_elasticsearch_input(monkeypatch, config):
    relations = [{"host": "1.2.3.4", "port": 1234}]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)

    telegraf.elasticsearch_input("test")
    expected = """
[[inputs.elasticsearch]]
  servers = ["http://1.2.3.4:1234"]
"""
    assert configs_dir().join("elasticsearch.conf").read().strip() == expected.strip()


def test_elasticsearch_input_no_relations(monkeypatch):
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    telegraf.elasticsearch_input("test")
    assert not configs_dir().join("elasticsearch.conf").exists()


def test_memcached_input(monkeypatch, config):
    relations = [{"host": "1.2.3.4", "port": 1234}]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.memcached_input("test")
    expected = """
[[inputs.memcached]]
  servers = ["1.2.3.4:1234"]
"""
    assert configs_dir().join("memcached.conf").read().strip() == expected.strip()


def test_memcached_input_no_relations(monkeypatch):
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    telegraf.memcached_input("test")
    assert not configs_dir().join("memcached.conf").exists()


def test_mongodb_input(monkeypatch, config):
    relations = [{"private-address": "1.2.3.4", "port": 1234}]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.mongodb_input("test")
    expected = """
[[inputs.mongodb]]
  servers = ["1.2.3.4:1234"]
"""
    assert configs_dir().join("mongodb.conf").read().strip() == expected.strip()


def test_mongodb_input_no_relations(monkeypatch):
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    telegraf.mongodb_input("test")
    assert not configs_dir().join("mongodb.conf").exists()


def test_haproxy_input(monkeypatch, config):
    relations = [
        {
            "private-address": "1.2.3.4",
            "port": 1234,
            "user": "foo",
            "password": "bar",
            "enabled": "True",
        }
    ]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    monkeypatch.setattr(telegraf.hookenv, "unit_private_ip", lambda: "1.2.3.4")
    telegraf.haproxy_input("test")
    expected = """
[[inputs.haproxy]]
  servers = ["http://foo:bar@1.2.3.4:1234"]
"""
    assert configs_dir().join("haproxy.conf").read().strip() == expected.strip()

    relations[0]["listener-address"] = "127.0.0.1"
    telegraf.haproxy_input("test")
    expected = """
[[inputs.haproxy]]
  servers = ["http://foo:bar@127.0.0.1:1234"]
"""
    assert configs_dir().join("haproxy.conf").read().strip() == expected.strip()


def test_haproxy_input_no_relations(monkeypatch):
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    telegraf.haproxy_input("test")
    assert not configs_dir().join("haproxy.conf").exists()


def test_haproxy_input_not_enabled(monkeypatch):
    relations = [
        {
            "private-address": "1.2.3.4",
            "port": 1234,
            "user": "foo",
            "password": "bar",
            "enabled": "False",
        }
    ]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.haproxy_input("test")
    assert not configs_dir().join("haproxy.conf").exists()


def test_apache_input(monkeypatch, config):
    relations = [{"__relid__": "apache:0", "private-address": "1.2.3.4"}]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    monkeypatch.setattr(telegraf.hookenv, "relation_set", lambda *a, **kw: None)
    telegraf.apache_input("test")
    expected = """
[[inputs.apache]]
  urls = ["http://1.2.3.4:8080/server-status?auto"]
"""
    assert configs_dir().join("apache.conf").read().strip() == expected.strip()


def test_apache_input_no_relations(monkeypatch):
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    telegraf.apache_input("test")
    assert not configs_dir().join("apache.conf").exists()


def test_exec_input(mocker, monkeypatch):
    interface = mocker.Mock(spec=RelationBase)
    interface.commands = mocker.Mock()
    command = {
        "commands": ["/srv/bin/test.sh", "/bin/true"],
        "data_format": "json",
        "timeout": "5s",
        "run_on_this_unit": True,
    }
    interface.commands.return_value = [command.copy()]
    telegraf.exec_input(interface)
    expected = """
[[inputs.exec]]
  commands = ['/srv/bin/test.sh', '/bin/true']
  data_format = "json"
  timeout = "5s"
"""
    assert configs_dir().join("exec.conf").read().strip() == expected.strip()
    # add a second relation/command set
    interface.commands.return_value = [command.copy(), command.copy()]
    telegraf.exec_input(interface)
    expected = expected + expected
    assert configs_dir().join("exec.conf").read().strip() == expected.strip()


def test_exec_input_with_tags(mocker, monkeypatch):
    interface = mocker.Mock(spec=RelationBase)
    interface.commands = mocker.Mock()
    commands = [
        {
            "commands": ["/srv/bin/test.sh", "/bin/true"],
            "data_format": "json",
            "timeout": "5s",
            "run_on_this_unit": True,
            "tags": {"test": "test"},
        }
    ]
    interface.commands.return_value = commands
    telegraf.exec_input(interface)
    expected = """
[[inputs.exec]]
  commands = ['/srv/bin/test.sh', '/bin/true']
  data_format = "json"
  timeout = "5s"
  [inputs.exec.tags]
    test = "test"
"""
    assert configs_dir().join("exec.conf").read().strip() == expected.strip()


def test_exec_input_no_leader(mocker, monkeypatch):
    interface = mocker.Mock(spec=RelationBase)
    interface.commands = mocker.Mock()
    commands = [
        {
            "commands": ["/srv/bin/test.sh", "/bin/true"],
            "data_format": "json",
            "timeout": "5s",
            "run_on_this_unit": False,
        }
    ]
    interface.commands.return_value = commands
    telegraf.exec_input(interface)
    assert not configs_dir().join("exec.conf").exists()


def test_exec_input_all_units(mocker, monkeypatch):
    interface = mocker.Mock(spec=RelationBase)
    interface.commands = mocker.Mock()
    commands = [
        {
            "commands": ["/srv/bin/test.sh", "/bin/true"],
            "data_format": "json",
            "timeout": "5s",
            "run_on_this_unit": True,
        }
    ]
    interface.commands.return_value = commands
    telegraf.exec_input(interface)
    expected = """
[[inputs.exec]]
  commands = ['/srv/bin/test.sh', '/bin/true']
  data_format = "json"
  timeout = "5s"
"""
    assert configs_dir().join("exec.conf").read().strip() == expected.strip()


def test_exec_input_leader_no_more(mocker, monkeypatch):
    interface = mocker.Mock(spec=RelationBase)
    interface.commands = mocker.Mock()
    command = {
        "commands": ["/srv/bin/test.sh", "/bin/true"],
        "data_format": "json",
        "timeout": "5s",
        "run_on_this_unit": True,
    }
    interface.commands.return_value = [command.copy()]
    telegraf.exec_input(interface)
    expected = """
[[inputs.exec]]
  commands = ['/srv/bin/test.sh', '/bin/true']
  data_format = "json"
  timeout = "5s"
"""
    assert configs_dir().join("exec.conf").read().strip() == expected.strip()
    # add a second relation/command set
    new_command = command.copy()
    new_command["run_on_this_unit"] = False
    interface.commands.return_value = [new_command.copy()]
    telegraf.exec_input(interface)
    assert not configs_dir().join("exec.conf").exists()


def test_exec_input_departed(mocker, monkeypatch):
    configs_dir().join("exec.conf").write("empty")
    relations = [1]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.exec_input_departed()
    assert configs_dir().join("exec.conf").exists()
    relations.pop()
    telegraf.exec_input_departed()
    assert not configs_dir().join("exec.conf").exists()


def test_influxdb_api_output(monkeypatch, config):
    relations = [
        {"hostname": "1.2.3.4", "port": 1234, "user": "foo", "password": "bar"}
    ]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.influxdb_api_output("test")
    expected = render(
        source="influxdb-api.conf.tmpl",
        target=None,
        templates_dir=telegraf.get_templates_dir(),
        context={
            "username": "foo",
            "password": "bar",
            "urls": '["http://1.2.3.4:1234"]',
        },
    )
    assert configs_dir().join("influxdb-api.conf").read().strip() == expected.strip()


def test_prometheus_client_output(mocker, monkeypatch, config):
    config["prometheus_output_port"] = ""  # Not configured globally
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: None)

    relation_ids = ["prometheus-client:0"]
    related_units = ["prometheus/0"]
    monkeypatch.setattr(telegraf.hookenv, "relation_ids", lambda r: relation_ids)
    monkeypatch.setattr(telegraf.hookenv, "related_units", lambda r: related_units)

    network_get_primary_address = mocker.patch.object(
        telegraf.hookenv, "network_get_primary_address", return_value="foo"
    )

    mocker.patch.object(
        telegraf.hookenv, "network_get", return_value={"ingress_addresses": ["1.2.3.4"]}
    )
    mocker.patch.object(telegraf.hookenv, "relation_get", return_value="5.6.7.8/32")

    interface = mocker.Mock(spec=RelationBase)
    interface.configure = mocker.Mock()

    # Test that telegraf will allow scraping from related prometheus, and from ranges in
    # prometheus_ip_range
    config["prometheus_ip_range"] = "127.0.0.1/32,10.0.0.0/8"
    telegraf.configure_prometheus_client_with_relation(interface)
    expected = """
    [[outputs.prometheus_client]]
  listen = ":9103"
  ip_range = ['127.0.0.1/32', '10.0.0.0/8', '5.6.7.8/32']
"""
    assert (
        configs_dir().join("prometheus_client.conf").read().strip() == expected.strip()
    )
    # prometheus_ip_range is empty, test that all IPs are allowed, not just the related
    # prometheus
    # prometheus_output_port is also empty, we should get 9103 by default
    network_get_primary_address.assert_called_once_with("prometheus-client")
    interface.configure.assert_called_once_with(
        "9103", hostname="foo", private_address="foo"
    )
    config["prometheus_ip_range"] = ""
    telegraf.configure_prometheus_client_with_relation(interface)
    expected = """
    [[outputs.prometheus_client]]
  listen = ":9103"
  ip_range = []
"""
    assert (
        configs_dir().join("prometheus_client.conf").read().strip() == expected.strip()
    )


def test_get_prometheus_ip_range_with_bad_data(mocker, monkeypatch, config):
    # if prometheus_ip_range is populated with an invalid IP, confirm we block
    config["prometheus_ip_range"] = "1.2.3.4, 5.6.7.8, banana"
    # we expect to raise this specific exception (and the hook to therefore fail)
    with pytest.raises(telegraf.InvalidPrometheusIPRange):
        telegraf.get_prometheus_ip_range()


def test_get_prometheus_ip_range_with_good_data(mocker, monkeypatch, config):
    # if prometheus_ip_range is populated with valid IPs, confirm we don't block
    status_set = mocker.patch("reactive.telegraf.hookenv.status_set")
    config["prometheus_ip_range"] = "192.168.0.0/24, 10.0.0.0/8"

    our_range = telegraf.get_prometheus_ip_range()
    # confirm we didn't block
    assert not status_set.called
    # and confirm we got the expected range back
    expected_range = ["192.168.0.0/24", "10.0.0.0/8"]
    assert our_range == expected_range


def test_prometheus_client_output_departed(mocker, monkeypatch, config):
    configs_dir().join("prometheus_client.conf").write("empty")
    relations = [1]
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: relations)
    telegraf.prometheus_client_departed()
    assert configs_dir().join("prometheus_client.conf").exists()
    relations.pop()
    telegraf.prometheus_client_departed()
    assert not configs_dir().join("prometheus_client.conf").exists()


# Integration tests (kind of)
def test_basic_config(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    check_prometheus_port = mocker.patch("reactive.telegraf.check_prometheus_port")
    set_flag("apt.installed.telegraf")
    assert not base_dir().join("telegraf.conf").exists()
    bus.dispatch()
    assert "telegraf.configured" in bus.get_states().keys()
    assert base_dir().join("telegraf.conf").exists()
    service_restart.assert_called_once_with(telegraf.DEB_SERVICE)
    check_prometheus_port.assert_called_once_with("prometheus_output", 9103)


def test_config_no_prometheus(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    check_prometheus_port = mocker.patch("reactive.telegraf.check_prometheus_port")
    config["prometheus_output_port"] = ""
    set_flag("apt.installed.telegraf")
    assert not base_dir().join("telegraf.conf").exists()
    bus.dispatch()
    assert "telegraf.configured" in bus.get_states().keys()
    assert base_dir().join("telegraf.conf").exists()
    service_restart.assert_called_once_with(telegraf.DEB_SERVICE)
    assert not check_prometheus_port.called


def test_config_changed_apt(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    mocker.patch("reactive.telegraf.check_prometheus_port")
    set_flag("apt.installed.telegraf")
    set_flag("config.changed")
    bus.dispatch()
    service_restart.assert_called_once_with(telegraf.DEB_SERVICE)


def test_config_changed_extra_options(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    mocker.patch("reactive.telegraf.check_prometheus_port")
    set_flag("apt.telegraf.installed")
    set_flag("plugins.haproxy.configured")
    config.save()
    config.load_previous()
    config["extra_options"] = yaml.dump({"inputs": {"haproxy": {"timeout": 10}}})
    set_flag("apt.installed.telegraf")
    set_flag("config.changed")
    bus.dispatch()
    assert "plugins.haproxy.configured" not in bus.get_states().keys()
    service_restart.assert_called_once_with(telegraf.DEB_SERVICE)


def test_config_changed_extra_plugins(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    mocker.patch("reactive.telegraf.check_prometheus_port")
    set_flag("apt.telegraf.installed")
    assert not configs_dir().join("extra_plugins.conf").exists()
    config.save()
    config.load_previous()
    # once the config is saved, change it. This will also trigger a service
    # restart
    config[
        "extra_plugins"
    ] = """[[inputs.foo]]
    some_option = "http://foo.bar.com"
[[outputs.baz]]
    option = "enabled"
    """
    set_flag("apt.installed.telegraf")
    set_flag("config.changed")
    bus.dispatch()
    assert configs_dir().join("extra_plugins.conf").exists()
    assert configs_dir().join("extra_plugins.conf").read() == config["extra_plugins"]
    service_restart.assert_called_with(telegraf.DEB_SERVICE)


def test_restart_on_output_plugin_relation_departed(mocker, monkeypatch, config):
    service_restart = mocker.patch("reactive.telegraf.host.service_restart")
    monkeypatch.setattr(telegraf.host, "service_running", lambda n: True)
    monkeypatch.setattr(telegraf.hookenv, "relations_of_type", lambda n: [])
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: None)
    relation_ids = ["prometheus-client:0"]
    monkeypatch.setattr(telegraf.hookenv, "relation_ids", lambda r: relation_ids)
    mocker.patch.object(
        telegraf.hookenv, "network_get", return_value={"ingress_addresses": ["1.2.3.4"]}
    )
    config["prometheus_output_port"] = ""
    bus.discover()
    set_flag("telegraf.installed")
    set_flag("snap.telegraf.installed")
    set_flag("telegraf.configured")
    interface = mocker.Mock(spec=RelationBase)
    interface.configure = mocker.Mock()
    telegraf.configure_prometheus_client_with_relation(interface)
    assert configs_dir().join("prometheus_client.conf").exists()
    # dispatch, file should be gone and telegraf restarted.
    bus.dispatch()
    assert not configs_dir().join("prometheus_client.conf").exists()
    service_restart.assert_called_with(telegraf.DEB_SERVICE)


def test_upgrade_charm(mocker):
    list_plugins = mocker.patch("reactive.telegraf.list_supported_plugins")
    clear_flag = mocker.patch("reactive.telegraf.clear_flag")
    list_plugins.return_value = ["foo", "bar"]
    telegraf.upgrade_charm()
    clear_flag.assert_any_call("telegraf.configured")
    clear_flag.assert_any_call("telegraf.apt.configured")
    clear_flag.assert_any_call("telegraf.snap.configured")
    clear_flag.assert_any_call("plugins.foo.configured")
    clear_flag.assert_any_call("plugins.bar.configured")
    clear_flag.assert_any_call("extra_plugins.configured")
    clear_flag.assert_any_call("grafana.configured")
    assert clear_flag.call_count == 7


def test_get_socket_listener_port(monkeypatch, config):
    config["socket_listener_port"] = ""
    assert telegraf.get_socket_listener_port() == 8094
    config["socket_listener_port"] = "9090"
    assert telegraf.get_socket_listener_port() == 9090


def test_socket_listener_config(monkeypatch, config):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    monkeypatch.setattr(telegraf.hookenv, "close_port", lambda p: open_ports.remove(p))
    config["prometheus_output_port"] = "default"
    telegraf.configure_telegraf()
    expected = 'service_address = "tcp://127.0.0.1:8094"'
    config_file = base_dir().join("telegraf.d", "socket_listener.conf")
    assert expected in config_file.read()


def test_inputs_internal_memstats(monkeypatch, config):
    open_ports = set()
    monkeypatch.setattr(telegraf.hookenv, "open_port", lambda p: open_ports.add(p))
    monkeypatch.setattr(telegraf.hookenv, "close_port", lambda p: open_ports.remove(p))
    config["disabled_plugins"] = ""  # ensure no metrics are initially disabled
    telegraf.configure_telegraf()
    config_file = base_dir().join("telegraf.conf")
    expected = """
[[inputs.internal]]
  collect_memstats = true
"""
    assert expected in config_file.read()


class TestGetHostnameLabel:
    @patch("reactive.telegraf.get_remote_unit_name")
    @patch("charmhelpers.core.hookenv.config")
    def test_legacy(self, config, get_remote_unit_name):
        config.return_value = dict(hostname="UNIT_NAME")
        get_remote_unit_name.return_value = "remote/10"
        assert telegraf.get_hostname_label() == "remote-10"

    @patch("reactive.telegraf.get_remote_unit_name")
    @patch("charmhelpers.core.hookenv.config")
    def test_juju1(self, config, get_remote_unit_name):
        get_remote_unit_name.return_value = "remote/10"
        env = dict(JUJU_ENV_NAME="env_name", JUJU_ENV_UUID="env_uuid")
        config.return_value = dict(hostname="{uuid}:{model}:{unit}")
        with patch.dict(os.environ, env):
            assert telegraf.get_hostname_label() == "env_uuid:env_name:remote-10"

    @patch("reactive.telegraf.get_remote_unit_name")
    @patch("charmhelpers.core.hookenv.config")
    def test_juju2(self, config, get_remote_unit_name):
        get_remote_unit_name.return_value = "remote/10"
        env = dict(JUJU_MODEL_NAME="model_name", JUJU_MODEL_UUID="model_uuid")
        config.return_value = dict(hostname="{uuid}:{model}:{unit}")
        with patch.dict(os.environ, env):
            assert telegraf.get_hostname_label() == "model_uuid:model_name:remote-10"


class TestTelegrafNagios:
    @mock.patch("charmhelpers.contrib.charmsupport.nrpe.NRPE")
    @mock.patch("charmhelpers.contrib.charmsupport.nrpe.get_nagios_hostname")
    def test_update_nrpe_config(self, mock_get_nagios_hostname, mock_nrpe, *args):
        mock_get_nagios_hostname.return_value = "testunit"
        svc = object()
        telegraf.configure_nagios(svc)
        calls = [
            call(hostname="testunit", primary=False),
            call().add_check(
                "telegraf_http",
                "Telegraf HTTP check",
                "check_http -I 127.0.0.1 -p 9103 -u " "/metrics",
            ),
            call().write(),
        ]
        mock_nrpe.assert_has_calls(calls)


class TestGrafanaDashboard:
    @patch("reactive.telegraf.render_custom")
    @patch("reactive.telegraf.endpoint_from_flag")
    @patch("reactive.telegraf.hookenv")
    @patch("reactive.telegraf.set_flag")
    @patch("reactive.telegraf.json")
    @patch("hashlib.md5")
    def test_register_grafana_dashboard(
        self,
        mock_md5,
        mock_json,
        mock_set_flag,
        mock_hookenv,
        mock_endpoint_from_flag,
        mock_render,
    ):
        expected_datasource = "my_prometheus"
        fake_config = dict(prometheus_datasource=expected_datasource)
        expected_dashboard_context = dict(
            datasource="{} - Juju generated source".format(expected_datasource),
            bonds_enabled=True,
            bcache_enabled=True,
            conntrack_enabled=True,
        )

        mock_grafana = MagicMock()
        mock_dashboard_dict = MagicMock()
        mock_rendered_content = MagicMock()

        mock_md5.return_value = MagicMock()
        mock_endpoint_from_flag.return_value = mock_grafana
        mock_hookenv.config.return_value = fake_config
        mock_json.loads.return_value = mock_dashboard_dict
        mock_render.return_value = mock_rendered_content

        telegraf.register_grafana_dashboard()

        mock_render.assert_called_once_with(
            source=telegraf.GRAFANA_DASHBOARD_TELEGRAF_FILE_NAME,
            render_context=expected_dashboard_context,
            variable_start_string="<<",
            variable_end_string=">>",
        )
        mock_grafana.register_dashboard.assert_called_once_with(
            name=telegraf.GRAFANA_DASHBOARD_NAME, dashboard=mock_dashboard_dict
        )
        mock_set_flag.assert_called_once_with("grafana.configured")

    @patch("reactive.telegraf.clear_flag")
    def test_unregister_grafana_dashboard(self, mock_clear_flag):
        telegraf.unregister_grafana_dashboard()
        mock_clear_flag.assert_called_with("grafana.configured")

    @patch("reactive.telegraf.endpoint_from_flag")
    @patch("reactive.telegraf.clear_flag")
    @patch("reactive.telegraf.hookenv")
    def test_grafana_dashboard_import_failed(
        self, mock_hookenv, mock_clear_flag, mock_endpoint_from_flag
    ):
        test_failed_imports = [
            MagicMock(name="import1", reason="some_reason1"),
            MagicMock(name="import2", reason="some_reason2"),
        ]

        mock_grafana = MagicMock()
        mock_grafana.failed_imports = test_failed_imports
        mock_endpoint_from_flag.return_value = mock_grafana

        telegraf.grafana_dashboard_import_failed()

        assert mock_hookenv.log.call_count == 2
        mock_clear_flag.assert_called_once_with("grafana.configured")


@pytest.mark.parametrize("install_method", ["deb", "snap"])
@mock.patch("reactive.telegraf.render")
def test_render_sudoers_file(render_mock, install_method):
    if install_method == "deb":
        context = {"telegraf_user": "telegraf"}
    elif install_method == "snap":
        context = {"telegraf_user": "snap_daemon"}
    filename = "test"
    template = "{}.tmpl".format(filename)
    target_mode = 0o640
    telegraf.render_sudoers_file(filename)
    render_mock.assert_called_with(
        source=template,
        templates_dir=os.path.join(telegraf.get_templates_dir(), "sudoers"),
        target=os.path.join(telegraf.SUDOERS_DIR, filename),
        context=context,
        perms=target_mode,
    )


@mock.patch("os.unlink")
@mock.patch("os.path.isfile")
def test_remove_sudoers_file(os_path_isfile_mock, os_unlink_mock):
    filename = "test"
    path = os.path.join(telegraf.SUDOERS_DIR, filename)

    # Test with non-existent file
    telegraf.remove_sudoers_file(filename)
    os_path_isfile_mock.assert_called_with(path)

    # Test with existent file
    def os_path_isfile_mock():
        return True

    telegraf.remove_sudoers_file(filename)
    os_unlink_mock.assert_called_with(path)


def test_collect_iptables_metrics(monkeypatch, config):
    config["collect_iptables_metrics"] = True
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    telegraf.configure_telegraf()

    expected = """
[[inputs.iptables]]
use_sudo = true
use_lock = true
binary = "iptables"
table = "filter"
chains = [ "INPUT", "OUTPUT", "FORWARD" ]

[[inputs.iptables]]
use_sudo = true
use_lock = true
binary = "iptables"
table = "nat"
chains = [ "INPUT", "OUTPUT", "PREROUTING", "POSTROUTING" ]

[[inputs.iptables]]
use_sudo = true
use_lock = true
binary = "iptables"
table = "raw"
chains = [ "OUTPUT", "PREROUTING" ]
"""
    config_file = base_dir().join("telegraf.conf")
    assert expected in config_file.read()


def test_collect_smart_metrics(monkeypatch, config):
    config["collect_smart_metrics"] = True
    monkeypatch.setattr(telegraf, "check_prometheus_port", lambda a, b: None)
    telegraf.configure_telegraf()

    expected = """
[[inputs.smart]]
use_sudo = true
"""
    config_file = base_dir().join("telegraf.conf")
    assert expected in config_file.read()
